# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, Optional

from torch import Tensor

from mmtrack.models.mot import BaseMultiObjectTracker
from mmtrack.registry import MODELS
from mmtrack.utils import OptConfigType, OptMultiConfig, SampleList


@MODELS.register_module()
class MaskTrackRCNN(BaseMultiObjectTracker):
    """Video Instance Segmentation.

    This video instance segmentor is the implementation of`MaskTrack R-CNN
    <https://arxiv.org/abs/1905.04804>`_.

    Args:
        detector (dict): Configuration of detector. Defaults to None.
        track_head (dict): Configuration of track head. Defaults to None.
        tracker (dict): Configuration of tracker. Defaults to None.
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
        init_cfg (dict or list[dict]): Configuration of initialization.
            Defaults to None.
    """

    def __init__(self,
                 detector: Optional[dict] = None,
                 track_head: Optional[dict] = None,
                 tracker: Optional[dict] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super().__init__(data_preprocessor, init_cfg)

        if detector is not None:
            self.detector = MODELS.build(detector)
        assert hasattr(self.detector, 'roi_head'), \
            'MaskTrack R-CNN only supports two stage detectors.'

        if track_head is not None:
            self.track_head = MODELS.build(track_head)
        if tracker is not None:
            self.tracker = MODELS.build(tracker)

    def loss(self, inputs: Dict[str, Tensor], data_samples: SampleList,
             **kwargs) -> dict:
        """Calculate losses from a batch of inputs and data samples.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size.The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            dict: A dictionary of loss components.
        """
        # modify the inputs shape to fit mmdet
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(1) == 1, \
            'MaskTrackRCNN can only have 1 key frame and 1 reference frame.'
        img = img[:, 0]

        ref_img = inputs['ref_img']
        assert ref_img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert ref_img.size(1) == 1, \
            'MaskTrackRCNN can only have 1 key frame and 1 reference frame.'
        ref_img = ref_img[:, 0]

        x = self.detector.extract_feat(img)
        ref_x = self.detector.extract_feat(ref_img)

        losses = dict()

        # RPN forward and loss
        if self.detector.with_rpn:
            proposal_cfg = self.detector.train_cfg.get(
                'rpn_proposal', self.detector.test_cfg.rpn)
            rpn_data_samples = copy.deepcopy(data_samples)
            rpn_losses, rpn_results_list = self.detector.rpn_head.\
                loss_and_predict(x,
                                 rpn_data_samples,
                                 proposal_cfg=proposal_cfg,
                                 **kwargs)
            # avoid get same name with roi_head loss
            keys = rpn_losses.keys()
            for key in keys:
                if 'loss' in key and 'rpn' not in key:
                    rpn_losses[f'rpn_{key}'] = rpn_losses.pop(key)
            losses.update(rpn_losses)
        else:
            # TODO: Not support currently, should have a check at Fast R-CNN
            assert data_samples[0].get('proposals', None) is not None
            # use pre-defined proposals in InstanceData for the second stage
            # to extract ROI features.
            rpn_results_list = [
                data_sample.proposals for data_sample in data_samples
            ]

        losses_detect = self.detector.roi_head.loss(x, rpn_results_list,
                                                    data_samples, **kwargs)
        losses.update(losses_detect)

        losses_track = self.track_head.loss(x, ref_x, rpn_results_list,
                                            data_samples, **kwargs)
        losses.update(losses_track)

        return losses

    def predict(self,
                inputs: dict,
                data_samples: SampleList,
                rescale: bool = True,
                **kwargs) -> SampleList:
        """Test without augmentation.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W)
                encoding input images. Typically these should be mean centered
                and std scaled. The N denotes batch size. The T denotes the
                number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
                In test mode, T = 1 and there is only ``img`` and no
                ``ref_img``.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and 'metainfo'.
            rescale (bool, Optional): If False, then returned bboxes and masks
                will fit the scale of img, otherwise, returned bboxes and masks
                will fit the scale of original image shape. Defaults to True.

        Returns:
            SampleList: Tracking results of the input images.
            Each TrackDataSample usually contains ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(1) == 1, \
            'MaskTrackRCNN can only have 1 key frame.'
        img = img[:, 0]

        assert len(data_samples) == 1, \
            'MaskTrackRCNN only support 1 batch size per gpu for now.'

        metainfo = data_samples[0].metainfo
        frame_id = metainfo.get('frame_id', -1)
        if frame_id == 0:
            self.tracker.reset()

        x = self.detector.extract_feat(img)

        rpn_results_list = self.detector.rpn_head.predict(x, data_samples)
        det_results = self.detector.roi_head.predict(
            x, rpn_results_list, data_samples, rescale=rescale)
        assert len(det_results) == 1, 'Batch inference is not supported.'
        assert 'masks' in det_results[0], 'There are no mask results.'
        track_data_sample = data_samples[0]
        track_data_sample.pred_det_instances = \
            det_results[0].clone()

        pred_track_instances = self.tracker.track(
            model=self,
            img=img,
            feats=x,
            data_sample=track_data_sample,
            rescale=rescale,
            **kwargs)
        track_data_sample.pred_track_instances = pred_track_instances

        return [track_data_sample]



_base_ = ['../../_base_/default_runtime.py']

randomness = dict(seed=1, deterministic=False)

# model setting
model = dict(
    type='PrDiMP',
    data_preprocessor=dict(
        type='TrackDataPreprocessor',
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        bgr_to_rgb=True),
    backbone=dict(
        type='mmcls.ResNet',
        depth=50,
        num_stages=3,
        strides=(1, 2, 2),
        dilations=[1, 1, 1],
        out_indices=[1, 2],  # 0, 1, 2
        norm_cfg=dict(type='BN', requires_grad=True),
        norm_eval=False,
        init_cfg=dict(type='Pretrained', checkpoint='torchvision://resnet50')),
    cls_head=dict(
        type='PrDiMPClsHead',
        in_dim=1024,
        out_dim=512,
        filter_initializer=dict(
            type='FilterInitializer',
            filter_size=4,
            feature_dim=512,
            feature_stride=16),
        filter_optimizer=dict(
            type='PrDiMPFilterOptimizer',
            num_iters=5,
            feat_stride=16,
            init_step_length=1.0,
            init_filter_regular=0.05,
            gauss_sigma=0.9,
            alpha_eps=0.05,
            min_filter_regular=0.05,
            label_thres=0),
        loss_cls=dict(type='KLGridLoss'),
        locate_cfg=dict(
            no_target_min_score=0.04,
            distractor_thres=0.8,
            hard_neg_thres=0.5,
            target_neighborhood_scale=2.2,
            dispalcement_scale=0.8,
            update_scale_when_uncertain=True),
        update_cfg=dict(
            sample_memory_size=50,
            normal_lr=0.01,
            hard_neg_lr=0.02,
            init_samples_min_weight=0.25,
            train_skipping=20),
        optimizer_cfg=dict(
            init_update_iters=10, update_iters=2, hard_neg_iters=1),
        train_cfg=dict(
            feat_size=(18, 18),
            img_size=(288, 288),
            sigma_factor=0.05,
            end_pad_if_even=True,
            gauss_label_bias=0.,
            use_gauss_density=True,
            label_density_norm=True,
            label_density_threshold=0.,
            label_density_shrink=0,
            loss_weights=dict(cls_init=0.25, cls_iter=1., cls_final=0.25))),
    bbox_head=dict(
        type='IouNetHead',
        in_dim=(4 * 128, 4 * 256),
        pred_in_dim=(256, 256),
        pred_inter_dim=(256, 256),
        loss_bbox=dict(type='KLMCLoss'),
        bbox_cfg=dict(
            num_init_random_boxes=9,
            box_jitter_pos=0.1,
            box_jitter_sz=0.5,
            iounet_topk=3,
            box_refine_step_length=2.5e-3,
            box_refine_iter=10,
            max_aspect_ratio=6,
            box_refine_step_decay=1),
        train_cfg=dict(
            proposals_sigma=[(0.05, 0.05), (0.5, 0.5)],
            gt_bboxes_sigma=(0.05, 0.05),
            num_samples=128,
            add_first_bbox=False,
            loss_weights=dict(bbox=0.0025))),
    test_cfg=dict(
        img_sample_size=22 * 16,
        feature_stride=16,
        search_scale_factor=6,
        patch_max_scale_change=1.5,
        border_mode='inside_major',
        bbox_inside_ratio=0.2,
        init_aug_cfg=dict(
            augmentation=dict(
                fliplr=True,
                rotate=[10, -10, 45, -45],
                blur=[(3, 1), (1, 3), (2, 2)],
                relativeshift=[(0.6, 0.6), (-0.6, 0.6), (0.6, -0.6),
                               (-0.6, -0.6)],
                dropout=[0.2, 0.2]),
            aug_expansion_factor=2,
            random_shift_factor=1 / 3)))

train_pipeline = [
    dict(
        type='DiMPSampling',
        num_search_frames=3,
        num_template_frames=3,
        max_frame_range=200),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile', to_float32=True),
            dict(type='LoadTrackAnnotations', with_instance_id=False),
            dict(type='GrayAug', prob=0.05)
        ]),
    dict(
        type='SeqBboxJitter',
        center_jitter_factor=[3, 3, 3, 4.5, 4.5, 4.5],
        scale_jitter_factor=[0.25, 0.25, 0.25, 0.5, 0.5, 0.5],
        crop_size_factor=[5, 5, 5, 5, 5, 5]),
    dict(
        type='TransformBroadcaster',
        share_random_params=False,
        transforms=[
            dict(type='CropLikeDiMP', crop_size_factor=5, output_size=288),
            dict(type='BrightnessAug', jitter_range=0.2)
        ]),
    dict(type='PackTrackInputs', ref_prefix='search', num_template_frames=3)
]

data_root = 'data/'
# dataset settings
train_dataloader = dict(
    batch_size=16,
    num_workers=4,
    persistent_workers=True,
    sampler=dict(type='QuotaSampler', samples_per_epoch=60000),
    dataset=dict(
        type='RandomSampleConcatDataset',
        dataset_sampling_weights=[1, 1, 1, 1],
        datasets=[
            dict(
                type='GOT10kDataset',
                data_root=data_root,
                ann_file='GOT10k/annotations/got10k_train_vot_infos.txt',
                data_prefix=dict(img_path='GOT10k'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='LaSOTDataset',
                data_root=data_root,
                ann_file='LaSOT_full/annotations/lasot_train_infos.txt',
                data_prefix=dict(img_path='LaSOT_full/LaSOTBenchmark'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='TrackingNetDataset',
                chunks_list=[0, 1, 2, 3],
                data_root=data_root,
                ann_file='TrackingNet/annotations/trackingnet_train_infos.txt',
                data_prefix=dict(img_path='TrackingNet'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='SOTCocoDataset',
                data_root=data_root,
                ann_file='coco/annotations/instances_train2017.json',
                data_prefix=dict(img_path='coco/train2017'),
                pipeline=train_pipeline,
                test_mode=False)
        ]))

# runner loop
train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=50, val_begin=50, val_interval=1)

# learning policy
param_scheduler = dict(type='StepLR', step_size=15, gamma=0.2)

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=2e-4),
    paramwise_cfg=dict(
        custom_keys=dict(
            backbone=dict(lr_multi=0.1),
            classifier=dict(lr_multi=5),
            bbox_regressor=dict(lr_multi=5))))

# checkpoint saving
default_hooks = dict(
    checkpoint=dict(type='CheckpointHook', interval=10),
    logger=dict(type='LoggerHook', interval=50))



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine.structures import BaseDataElement, InstanceData


class TrackDataSample(BaseDataElement):
    """A data structure interface of MMTracking. They are used as interfaces
    between different components.

    The attributes in ``TrackDataSample`` are divided into several parts:

        - ``gt_instances``(InstanceData): Ground truth of instance annotations
            in key frames.
        - ``ignored_instances``(InstanceData): Instances to be ignored during
            training/testing in key frames.
        - ``proposals``(InstanceData): Region proposals used in two-stage
            detectors in key frames.
        - ``ref_gt_instances``(InstanceData): Ground truth of instance
            annotations in reference frames.
        - ``ref_ignored_instances``(InstanceData): Instances to be ignored
            during training/testing in reference frames.
        - ``ref_proposals``(InstanceData): Region proposals used in two-stage
            detectors in reference frames.
        - ``pred_det_instances``(InstanceData): Detection instances of model
            predictions in key frames.
        - ``pred_track_instances``(InstanceData): Tracking instances of model
            predictions in key frames.
    """

    # Typically used in key frames.
    @property
    def gt_instances(self) -> InstanceData:
        return self._gt_instances

    @gt_instances.setter
    def gt_instances(self, value: InstanceData):
        self.set_field(value, '_gt_instances', dtype=InstanceData)

    @gt_instances.deleter
    def gt_instances(self):
        del self._gt_instances

    # Typically used in key frames.
    @property
    def ignored_instances(self) -> InstanceData:
        return self._ignored_instances

    @ignored_instances.setter
    def ignored_instances(self, value: InstanceData):
        self.set_field(value, '_ignored_instances', dtype=InstanceData)

    @ignored_instances.deleter
    def ignored_instances(self):
        del self._ignored_instances

    # Typically used in key frames.
    @property
    def proposals(self) -> InstanceData:
        return self._proposals

    @proposals.setter
    def proposals(self, value: InstanceData):
        self.set_field(value, '_proposals', dtype=InstanceData)

    @proposals.deleter
    def proposals(self):
        del self._proposals

    # Typically denotes the detection results of key frame
    @property
    def pred_det_instances(self) -> InstanceData:
        return self._pred_det_instances

    @pred_det_instances.setter
    def pred_det_instances(self, value: InstanceData):
        self.set_field(value, '_pred_det_instances', dtype=InstanceData)

    @pred_det_instances.deleter
    def pred_det_instances(self):
        del self._pred_det_instances

    # Typically denotes the tracking results of key frame
    @property
    def pred_track_instances(self) -> InstanceData:
        return self._pred_track_instances

    @pred_track_instances.setter
    def pred_track_instances(self, value: InstanceData):
        self.set_field(value, '_pred_track_instances', dtype=InstanceData)

    @pred_track_instances.deleter
    def pred_track_instances(self):
        del self._pred_track_instances



# model settings
_base_ = ['./qdtrack_faster-rcnn_r101_fpn_8xb2-24e_lvis_test-tao.py']

model = dict(freeze_detector=True)

data_root = 'data/tao/'
train_dataloader = dict(
    batch_size=2,
    num_workers=2,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    batch_sampler=dict(type='mmdet.AspectRatioBatchSampler'),
    dataset=dict(
        dataset=dict(
            data_root=data_root,
            data_prefix=dict(img_path=''),
            load_as_video=True,
            metainfo=dict(CLASSES=(data_root + 'annotations/tao_classes.txt')),
            ann_file='annotations/train_482_classes.json',
            ref_img_sampler=dict(
                num_ref_imgs=1, frame_range=[-1, 1], method='uniform'))))

# learning policy
param_scheduler = [
    dict(
        type='LinearLR',
        start_factor=1.0 / 1000,
        by_epoch=False,
        begin=0,
        end=1000),
    dict(
        type='mmdet.MultiStepLR',
        begin=0,
        end=12,
        by_epoch=True,
        milestones=[8, 11])
]
# runtime settings
train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=12, val_begin=0, val_interval=1)

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.002, momentum=0.9, weight_decay=0.0001),
    clip_grad=None)

val_evaluator = dict(type='TAOMetric', metric=['tao_track_ap'])
test_evaluator = val_evaluator
load_from = None



# Copyright (c) OpenMMLab. All rights reserved.
import random
from collections import defaultdict
from typing import Dict, List, Optional

import numpy as np
from mmcv.transforms import BaseTransform
from mmengine.logging import print_log

from mmtrack.registry import TRANSFORMS


@TRANSFORMS.register_module()
class DiMPSampling(BaseTransform):
    """DiMP-style sampling. It's firstly used in `DiMP.

    <https://arxiv.org/abs/1904.07220>`_.

    Required Keys:

    - img_paths
    - frame_ids
    - video_id
    - video_length
    - bboxes
    - instance_id (optional)
    - mask (optional)

    - seg_map_path (optional)

    Added Keys:

    - instances

      - bbox (np.float32)
      - bbox_label (np.int32)
      - frame_id (np.int32)
      - ignore_flag (np.bool)
      - img_path (str)

    Args:
        num_search_frames (int, optional): the number of search frames
        num_template_frames (int, optional): the number of template frames
        max_frame_range (list[int], optional): the max frame range of sampling
            a positive search image for the template image. Its length is equal
            to the number of extra templates, i.e., `num_template_frames`-1.
            Default length is 1.
        min_num_frames (int, optional): the min number of frames to be sampled.
    """

    def __init__(self,
                 num_search_frames: int = 3,
                 num_template_frames: int = 3,
                 max_frame_range: int = 200,
                 min_num_frames: int = 20):
        self.num_search_frames = num_search_frames
        self.num_template_frames = num_template_frames
        self.max_frame_range = max_frame_range
        self.min_num_frames = min_num_frames

    def random_sample_inds(self,
                           video_visibility: np.ndarray,
                           num_samples: int = 1,
                           frame_range: Optional[List] = None,
                           allow_invisible: bool = False,
                           force_invisible: bool = False) -> List[int]:
        """Random sampling a specific number of samples from the specified
        frame range of the video. It also considers the visibility of each
        frame.

        Args:
            video_visibility (np.ndarray): the visibility of each frame in the
                video.
            num_samples (int, optional): the number of samples. Defaults to 1.
            frame_range (list | None, optional): the frame range of sampling.
                Defaults to None.
            allow_invisible (bool, optional): whether to allow to get invisible
                samples. Defaults to False.
            force_invisible (bool, optional): whether to force to get invisible
                samples. Defaults to False.

        Returns:
            List[int]: The sampled frame indexes.
        """
        assert num_samples > 0
        if frame_range is None:
            frame_range = [0, len(video_visibility)]
        else:
            assert isinstance(frame_range, list) and len(frame_range) == 2
            frame_range[0] = max(0, frame_range[0])
            frame_range[1] = min(len(video_visibility), frame_range[1])

        video_visibility = np.asarray(video_visibility)
        visibility_in_range = video_visibility[frame_range[0]:frame_range[1]]
        # get indexes of valid samples
        if force_invisible:
            valid_inds = np.where(~visibility_in_range)[0] + frame_range[0]
        else:
            valid_inds = np.arange(
                *frame_range) if allow_invisible else np.where(
                    visibility_in_range)[0] + frame_range[0]

        # No valid samples
        if len(valid_inds) == 0:
            return [None] * num_samples

        return random.choices(valid_inds, k=num_samples)

    def sampling_frames(self, video_visibility: np.ndarray) -> List:
        """Sampling multiple template images and one search images in one
        video.

        Args:
            video_visibility (np.ndarray): the visibility of each frame in the
                video.

        Returns:
            List: the indexes of template and search images.
        """
        search_frame_inds = [None]
        gap_increase = 0
        if self.is_video_data:
            while search_frame_inds[0] is None:
                # first randomly sample two frames from a video
                base_frame_ind = self.random_sample_inds(
                    video_visibility,
                    num_samples=1,
                    frame_range=[
                        self.num_template_frames - 1,
                        len(video_visibility) - self.num_search_frames
                    ])

                prev_frame_inds = self.random_sample_inds(
                    video_visibility,
                    num_samples=self.num_template_frames - 1,
                    frame_range=[
                        base_frame_ind[0] - self.max_frame_range -
                        gap_increase, base_frame_ind[0]
                    ])

                if prev_frame_inds[0] is None:
                    gap_increase += 5
                    continue

                temp_frame_inds = base_frame_ind + prev_frame_inds
                search_frame_inds = self.random_sample_inds(
                    video_visibility,
                    num_samples=self.num_search_frames,
                    frame_range=[
                        temp_frame_inds[0] + 1, temp_frame_inds[0] +
                        self.max_frame_range + gap_increase
                    ])

                gap_increase += 5

            sampled_inds = temp_frame_inds + search_frame_inds
        else:
            sampled_inds = [0] * (
                self.num_template_frames + self.num_search_frames)

        return sampled_inds

    def prepare_data(self,
                     video_info: dict,
                     sampled_inds: List[int],
                     is_positive_pairs: bool = True,
                     results: Optional[dict] = None) -> Dict[str, List]:
        """Prepare sampled training data according to the sampled index.

        Args:
            video_info (dict): the video information. It contains the keys:
                ['bboxes', 'bboxes_isvalid', 'img_paths', 'frame_ids',
                'video_id', 'visible', 'video_length].
            sampled_inds (list[int]): the sampled frame indexes.
            is_positive_pairs (bool, optional): whether it's the positive
                pairs. Defaults to True.
            results (dict[list], optional): The prepared results which need to
                be updated. Defaults to None.

        Returns:
            Dict[str, List]: contains the information of sampled data.
        """
        if results is None:
            results = defaultdict(list)
        assert isinstance(results, dict)
        for frame_ind in sampled_inds:
            results['img_path'].append(video_info['img_paths'][frame_ind])
            results['frame_id'].append(video_info['frame_ids'][frame_ind])
            results['video_id'].append(video_info['video_id'])
            results['video_length'].append(video_info['video_length'])
            instance = [
                dict(
                    bbox=video_info['bboxes'][frame_ind],
                    bbox_label=np.array(is_positive_pairs, dtype=np.int32))
            ]
            results['instances'].append(instance)
        return results

    def transform(self,
                  pair_video_infos: List[dict]) -> Optional[Dict[str, List]]:
        """
        Args:
            pair_video_infos (list[dict]): contains two video infos. Each video
                info contains the keys: ['bboxes','bboxes_isvalid','filename',
                'frame_ids','video_id','visible'].

        Returns:
            Optional[Dict[str, List]]: contains the information of sampled
                data.
        """
        video_info, video_info_another = pair_video_infos
        self.is_video_data = len(video_info['frame_ids']) > 1 and len(
            video_info_another['frame_ids']) > 1
        enough_visible_frames = sum(video_info['visible']) > 2 * (
            self.num_search_frames + self.num_template_frames) and len(
                video_info['visible']) >= self.min_num_frames
        enough_visible_frames = enough_visible_frames or not \
            self.is_video_data

        if not enough_visible_frames:
            return None

        sampled_inds = np.array(self.sampling_frames(video_info['visible']))
        # the sizes of some bboxes may be zero, because extral templates may
        # get invalid bboxes.
        if not video_info['bboxes_isvalid'][sampled_inds].all():
            return None

        results = self.prepare_data(video_info, sampled_inds)

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'num_search_frames={self.num_search_frames}, '
        repr_str += f'num_template_frames={self.num_template_frames}, '
        repr_str += f'max_frame_range={self.max_frame_range})'
        repr_str += f'min_num_frames={self.min_num_frames})'
        return repr_str


@TRANSFORMS.register_module()
class TridentSampling(DiMPSampling):
    """Multitemplate-style sampling in a trident manner. It's firstly used in
    `STARK <https://arxiv.org/abs/2103.17154.>`_.

    The input in this transform is a list of dict. In each dict:

    Required Keys:

    - img_paths
    - frame_ids
    - video_id
    - video_length
    - bboxes
    - instance_id (optional)
    - mask (optional)

    - seg_map_path (optional)

    Added Keys:

    - instances

      - bbox (np.float32)
      - bbox_label (np.int32)
      - frame_id (np.int32)
      - ignore_flag (np.bool)
      - img_path (str)

    Args:
        num_search_frames (int, optional): the number of search frames
        num_template_frames (int, optional): the number of template frames
        max_frame_range (list[int], optional): the max frame range of sampling
            a positive search image for the template image. Its length is equal
            to the number of extra templates, i.e., `num_template_frames`-1.
            Default length is 1.
        cls_pos_prob (float, optional): the probility of sampling positive
            samples in classification training.
        train_cls_head (bool, optional): whether to train classification head.
        min_num_frames (int, optional): the min number of frames to be sampled.
    """

    def __init__(self,
                 num_search_frames: int = 1,
                 num_template_frames: int = 2,
                 max_frame_range: List[int] = [200],
                 min_num_frames: int = 20,
                 cls_pos_prob: float = 0.5,
                 train_cls_head: bool = False):
        assert num_template_frames >= 2
        assert len(max_frame_range) == num_template_frames - 1
        super().__init__(num_search_frames, num_template_frames,
                         max_frame_range, min_num_frames)
        self.train_cls_head = train_cls_head
        self.cls_pos_prob = cls_pos_prob

    def sampling_frames(self, video_visibility: np.ndarray) -> List[int]:
        """Sampling multiple template images and one search images in one
        video.

        Args:
            video_visibility (np.ndarray): the visibility of each frame in the
                video.

        Returns:
            List[int]: the indexes of template and search images.
        """
        extra_template_inds = [None]
        sampling_count = 0
        if self.is_video_data:
            while None in extra_template_inds:
                # first randomly sample two frames from a video
                template_ind, search_ind = self.random_sample_inds(
                    video_visibility, num_samples=2)

                # then sample the extra templates
                extra_template_inds = []
                for max_frame_range in self.max_frame_range:
                    # make the sampling range is near the template_ind
                    if template_ind >= search_ind:
                        min_ind, max_ind = search_ind, \
                            search_ind + max_frame_range
                    else:
                        min_ind, max_ind = search_ind - max_frame_range, \
                            search_ind
                    extra_template_index = self.random_sample_inds(
                        video_visibility,
                        num_samples=1,
                        frame_range=[min_ind, max_ind],
                        allow_invisible=False)[0]

                    extra_template_inds.append(extra_template_index)

                sampling_count += 1
                if sampling_count > 100:
                    print_log('-------Not sampling extra valid templates'
                              'successfully. Stop sampling and copy the'
                              'first template as extra templates-------')
                    extra_template_inds = [template_ind] * len(
                        self.max_frame_range)

            sampled_inds = [template_ind] + extra_template_inds + [search_ind]
        else:
            sampled_inds = [0] * (
                self.num_template_frames + self.num_search_frames)

        return sampled_inds

    def prepare_cls_data(self, video_info: dict, video_info_another: dict,
                         sampled_inds: List[int]) -> Dict[str, List]:
        """Prepare the sampled classification training data according to the
        sampled index.

        Args:
            video_info (dict): the video information. It contains the keys:
                ['bboxes', 'bboxes_isvalid', 'filename', 'frame_ids',
                'video_id', 'visible', 'video_length].
            video_info_another (dict): the another video information. It's only
                used to get negative samples in classification train. It
                contains the keys: ['bboxes','bboxes_isvalid','filename',
                'frame_ids','video_id','visible','video_length]].
            sampled_inds (list[int]): the sampled frame indexes.

        Returns:
            Dict[str, List]: contains the information of sampled data.
        """
        if random.random() < self.cls_pos_prob:
            results = self.prepare_data(
                video_info, sampled_inds, is_positive_pairs=True)
        else:
            results = self.prepare_data(
                video_info,
                sampled_inds[:self.num_template_frames],
                is_positive_pairs=False)

            if self.is_video_data:
                neg_search_ind = self.random_sample_inds(
                    video_info_another['bboxes_isvalid'], num_samples=1)
                # may not get valid negative sample in current video
                if neg_search_ind[0] is None:
                    return None
            else:
                neg_search_ind = [0]

            results = self.prepare_data(
                video_info_another,
                neg_search_ind,
                is_positive_pairs=False,
                results=results)

        return results

    def transform(self,
                  pair_video_infos: List[dict]) -> Optional[Dict[str, List]]:
        """
        Args:
            pair_video_infos (list[dict]): contains two video infos. Each video
                info contains the keys: ['bboxes','bboxes_isvalid','filename',
                'frame_ids','video_id','visible','video_length'].

        Returns:
            Optional[Dict[str, List]]: contains the information of sampled
                data. If not enough visible frames, return None.
        """
        video_info, video_info_another = pair_video_infos
        self.is_video_data = len(video_info['frame_ids']) > 1 and len(
            video_info_another['frame_ids']) > 1
        enough_visible_frames = sum(video_info['visible']) > 2 * (
            self.num_search_frames + self.num_template_frames) and len(
                video_info['visible']) >= self.min_num_frames
        enough_visible_frames = enough_visible_frames or not \
            self.is_video_data

        if not enough_visible_frames:
            return None

        sampled_inds = np.array(self.sampling_frames(video_info['visible']))
        # the sizes of some bboxes may be zero, because extral templates may
        # get invalid bboxes.
        if not video_info['bboxes_isvalid'][sampled_inds].all():
            return None

        if not self.train_cls_head:
            results = self.prepare_data(video_info, sampled_inds)
        else:
            results = self.prepare_cls_data(video_info, video_info_another,
                                            sampled_inds)

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'num_search_frames={self.num_search_frames}, '
        repr_str += f'num_template_frames={self.num_template_frames}, '
        repr_str += f'max_frame_range={self.max_frame_range}, '
        repr_str += f'cls_pos_prob={self.cls_pos_prob}, '
        repr_str += f'train_cls_head={self.train_cls_head}, '
        repr_str += f'min_num_frames={self.min_num_frames})'
        return repr_str


@TRANSFORMS.register_module()
class PairSampling(BaseTransform):
    """Pair-style sampling. It's used in `SiameseRPN++

    <https://arxiv.org/abs/1812.11703.>`_.

    The input in this transform is a list of dict. In each dict:

    Required Keys:

    - img_paths
    - frame_ids
    - video_id
    - video_length
    - bboxes
    - instance_id (optional)
    - mask (optional)

    - seg_map_path (optional)

    Added Keys:

    - instances

      - bbox (np.float32)
      - bbox_label (np.int32)
      - frame_id (np.int32)
      - ignore_flag (np.bool)
      - img_path (str)

    Args:
        frame_range (List(int) | int): The sampling range of search
            frames in the same video for template frame. Defaults to 5.
        pos_prob (float, optional):  The probility of sampling positive
            sample pairs. Defaults to 0.8.
        filter_template_img (bool, optional): If False, the template image will
            be in the sampling search candidates, otherwise, it is exclude.
            Defaults to False.
    """

    def __init__(self,
                 frame_range: int = 5,
                 pos_prob: float = 0.8,
                 filter_template_img: bool = False):
        assert pos_prob >= 0.0 and pos_prob <= 1.0
        if isinstance(frame_range, int):
            assert frame_range >= 0, 'frame_range can not be a negative value.'
            frame_range = [-frame_range, frame_range]
        elif isinstance(frame_range, list):
            assert len(frame_range) == 2, 'The length must be 2.'
            assert frame_range[0] <= 0 and frame_range[1] >= 0
            for i in frame_range:
                assert isinstance(i, int), 'Each element must be int.'
        else:
            raise TypeError('The type of frame_range must be int or list.')
        self.frame_range = frame_range
        self.pos_prob = pos_prob
        self.filter_template_img = filter_template_img

    def prepare_data(self,
                     video_info: dict,
                     sampled_inds: List[int],
                     is_positive_pairs: bool = False,
                     results: Optional[dict] = None) -> Dict[str, List]:
        """Prepare sampled training data according to the sampled index.

        Args:
            video_info (dict): the video information. It contains the keys:
                ['bboxes', 'bboxes_isvalid', 'img_paths', 'frame_ids',
                'video_id', 'visible', 'video_length].
            sampled_inds (list[int]): the sampled frame indexes.
            is_positive_pairs (bool, optional): whether it's the positive
                pairs. Defaults to False.
            results (dict[list], optional): The prepared results which need to
                be updated. Defaults to None.

        Returns:
            Dict[str, List]: contains the information of sampled data.
        """
        if results is None:
            results = defaultdict(list)
        assert isinstance(results, dict)
        for frame_ind in sampled_inds:
            results['img_path'].append(video_info['img_paths'][frame_ind])
            results['frame_id'].append(video_info['frame_ids'][frame_ind])
            results['video_id'].append(video_info['video_id'])
            results['video_length'].append(video_info['video_length'])
            instance = [
                dict(
                    bbox=video_info['bboxes'][frame_ind],
                    bbox_label=np.array(is_positive_pairs, dtype=np.int32))
            ]
            results['instances'].append(instance)
        return results

    def transform(self, pair_video_infos: List[dict]) -> dict:
        """
        Args:
            pair_video_infos (list[dict]): Containing the information of two
                videos. Each video information contains the keys:
                ['bboxes','bboxes_isvalid', 'img_paths', 'frame_ids',
                'video_id', 'visible', 'video_length'].

        Returns:
            dict: contains the information of sampled data.
        """
        video_info, video_info_another = pair_video_infos
        if len(video_info['frame_ids']) > 1 and len(
                video_info_another['frame_ids']) > 1:
            template_frame_ind = np.random.choice(len(video_info['frame_ids']))
            if self.pos_prob > np.random.random():
                left_ind = max(template_frame_ind + self.frame_range[0], 0)
                right_ind = min(template_frame_ind + self.frame_range[1],
                                len(video_info['frame_ids']))
                if self.filter_template_img:
                    ref_frames_inds = list(
                        range(left_ind, template_frame_ind)) + list(
                            range(template_frame_ind + 1, right_ind))
                else:
                    ref_frames_inds = list(range(left_ind, right_ind))
                search_frame_ind = np.random.choice(ref_frames_inds)
                results = self.prepare_data(
                    video_info, [template_frame_ind, search_frame_ind],
                    is_positive_pairs=True)
            else:
                search_frame_ind = np.random.choice(
                    len(video_info_another['frame_ids']))
                results = self.prepare_data(
                    video_info, [template_frame_ind], is_positive_pairs=False)
                results = self.prepare_data(
                    video_info_another, [search_frame_ind],
                    is_positive_pairs=False,
                    results=results)

        else:
            if self.pos_prob > np.random.random():
                results = self.prepare_data(
                    video_info, [0, 0], is_positive_pairs=True)
            else:
                results = self.prepare_data(
                    video_info, [0], is_positive_pairs=False)
                results = self.prepare_data(
                    video_info_another, [0],
                    is_positive_pairs=False,
                    results=results)

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'frame_range={self.frame_range}, '
        repr_str += f'pos_prob={self.pos_prob}, '
        repr_str += f'filter_template_img={self.filter_template_img})'
        return repr_str



_base_ = [
    './tracktor_faster-rcnn_r50-fpn_8xb2-4e_mot17halftrain'
    '_test-mot17halfval.py'
]

model = dict(
    detector=dict(
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/faster_rcnn/faster-rcnn_r50_fpn_4e_mot16-half_20210804_001054-73477869.pth'  # noqa: E501
        )),
    reid=dict(
        head=dict(num_classes=375),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/reid/reid_r50_6e_mot16_20210803_204826-1b3e3cfd.pth'  # noqa: E501
        )))

# dataloader
data_root = 'data/MOT16/'
val_dataloader = dict(
    dataset=dict(
        data_root=data_root,
        ann_file='annotations/half-val_cocoformat.json',
    ))
test_dataloader = val_dataloader



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import List, Tuple

from mmdet.datasets.api_wrappers import COCO
from mmengine.fileio import FileClient

from mmtrack.registry import DATASETS
from .api_wrappers import CocoVID
from .base_video_dataset import BaseVideoDataset


@DATASETS.register_module()
class ImagenetVIDDataset(BaseVideoDataset):
    """ImageNet VID dataset for video object detection."""

    METAINFO = {
        'CLASSES':
        ('airplane', 'antelope', 'bear', 'bicycle', 'bird', 'bus', 'car',
         'cattle', 'dog', 'domestic_cat', 'elephant', 'fox', 'giant_panda',
         'hamster', 'horse', 'lion', 'lizard', 'monkey', 'motorcycle',
         'rabbit', 'red_panda', 'sheep', 'snake', 'squirrel', 'tiger', 'train',
         'turtle', 'watercraft', 'whale', 'zebra')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from an annotation file named as ``self.ann_file``.
        Specifically, if self.load_as_video is True, it loads from the video
        annotation file. Otherwise, from the image annotation file.

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        if self.load_as_video:
            data_list, valid_data_indices = self._load_video_data_list()
        else:
            data_list, valid_data_indices = self._load_image_data_list()

        return data_list, valid_data_indices

    def _load_video_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from a video annotation file named as
        ``self.ann_file``.

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_path:
            coco = CocoVID(local_path)
        # The order of returned `cat_ids` will not
        # change with the order of the CLASSES
        self.cat_ids = coco.get_cat_ids(cat_names=self.metainfo['CLASSES'])
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.cat_img_map = copy.deepcopy(coco.cat_img_map)

        data_list = []
        valid_data_indices = []
        data_id = 0
        vid_ids = coco.get_vid_ids()

        for vid_id in vid_ids:
            img_ids = coco.get_img_ids_from_vid(vid_id)
            for img_id in img_ids:
                # load img info
                raw_img_info = coco.load_imgs([img_id])[0]
                raw_img_info['img_id'] = img_id
                raw_img_info['video_length'] = len(img_ids)

                # load ann info
                ann_ids = coco.get_ann_ids(
                    img_ids=[img_id], cat_ids=self.cat_ids)
                raw_ann_info = coco.load_anns(ann_ids)

                # load frames for training
                if self.test_mode:
                    assert not raw_img_info['is_vid_train_frame'], \
                        'is_vid_train_frame must be False in testing'
                    valid_data_indices.append(data_id)
                elif raw_img_info['is_vid_train_frame']:
                    valid_data_indices.append(data_id)

                # get data_info
                parsed_data_info = self.parse_data_info(
                    dict(raw_img_info=raw_img_info, raw_ann_info=raw_ann_info))
                data_list.append(parsed_data_info)
                data_id += 1
        assert len(
            valid_data_indices
        ) != 0, f"There is no frame for training in '{self.ann_file}'!"

        return data_list, valid_data_indices

    def _load_image_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from an image annotation file named as
        ``self.ann_file``.

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_path:
            coco = COCO(local_path)
        # The order of returned `cat_ids` will not
        # change with the order of the CLASSES
        self.cat_ids = coco.get_cat_ids(cat_names=self.metainfo['CLASSES'])
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.cat_img_map = copy.deepcopy(coco.cat_img_map)

        img_ids = coco.get_img_ids()
        data_id = 0
        valid_data_indices = []
        data_list = []
        total_ann_ids = []
        for img_id in img_ids:
            raw_img_info = coco.load_imgs([img_id])[0]
            raw_img_info['img_id'] = img_id

            ann_ids = coco.get_ann_ids(img_ids=[img_id])
            raw_ann_info = coco.load_anns(ann_ids)
            total_ann_ids.extend(ann_ids)

            # load images for training
            if raw_img_info['is_vid_train_frame']:
                valid_data_indices.append(data_id)

            parsed_data_info = self.parse_data_info(
                dict(raw_img_info=raw_img_info, raw_ann_info=raw_ann_info))
            data_list.append(parsed_data_info)
            data_id += 1
        assert len(set(total_ann_ids)) == len(
            total_ann_ids
        ), f"Annotation ids in '{self.ann_file}' are not unique!"

        return data_list, valid_data_indices



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Tuple

import torch
from mmdet.structures.bbox import bbox_overlaps
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.structures import TrackDataSample
from ..task_modules.track import embed_similarity
from .base_tracker import BaseTracker


@MODELS.register_module()
class QuasiDenseTAOTracker(BaseTracker):
    """Tracker for Quasi-Dense Tracking Method with TAO Dataset.

    Args:
        init_score_thr (float): The cls_score threshold to
            initialize a new tracklet. Defaults to 0.8.
        obj_score_thr (float): The cls_score threshold to
            update a tracked tracklet. Defaults to 0.5.
        match_score_thr (float): The match threshold. Defaults to 0.5.
        memo_frames (int): The most frames in a track memory.
            Defaults to 10.
        memo_momentum (float): The momentum value for embeds updating.
            Defaults to 0.8.
        momentum_obj_score (float): The momentum value for object score
            updating. Default to 0.5.
        obj_score_diff_thr (float): The threshold for object score
            difference for adjacent detection in the same track.
        nms_conf_thr (float): The nms threshold for confidence.
            Defaults to 0.5.
        distractor_nms_thr (float): The nms threshold for inter-class.
            Defaults to 0.3.
        distractor_score_thr (float): The threshold for distractor.
            Defaults to 0.5.
        match_metric (str): The match metric. Defaults to 'bisoftmax'.
        match_with_cosine (bool): If True, match score contains cosine
            similarity. Default to True.
    """

    def __init__(self,
                 init_score_thr: float = 0.0001,
                 obj_score_thr: float = 0.0001,
                 match_score_thr: float = 0.5,
                 memo_frames: int = 10,
                 memo_momentum: float = 0.8,
                 momentum_obj_score: float = 0.5,
                 obj_score_diff_thr: float = 1.0,
                 distractor_nms_thr: float = 0.3,
                 distractor_score_thr: float = 0.5,
                 match_metric: str = 'bisoftmax',
                 match_with_cosine: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        self.init_score_thr = init_score_thr
        self.obj_score_thr = obj_score_thr
        self.match_score_thr = match_score_thr

        self.memo_frames = memo_frames
        self.memo_momentum = memo_momentum
        self.momentum_obj_score = momentum_obj_score
        self.obj_score_diff_thr = obj_score_diff_thr
        self.distractor_nms_thr = distractor_nms_thr
        self.distractor_score_thr = distractor_score_thr
        assert match_metric in ['bisoftmax', 'cosine']
        self.match_metric = match_metric
        self.match_with_cosine = match_with_cosine

        self.num_tracks = 0
        self.tracks = dict()

    def reset(self):
        """Reset the buffer of the tracker."""
        self.num_tracks = 0
        self.tracks = dict()

    def update(self, ids: Tensor, bboxes: Tensor, labels: Tensor,
               embeds: Tensor, scores: Tensor, frame_id: int) -> None:
        """Tracking forward function.

        Args:
            ids (Tensor): of shape(N, ).
            bboxes (Tensor): of shape (N, 5).
            embeds (Tensor): of shape (N, 256).
            labels (Tensor): of shape (N, ).
            scores (Tensor): of shape (N, ).
            frame_id (int): The id of current frame, 0-index.
        """
        tracklet_inds = ids > -1

        # update memo
        for id, bbox, embed, label, score in zip(ids[tracklet_inds],
                                                 bboxes[tracklet_inds],
                                                 embeds[tracklet_inds],
                                                 labels[tracklet_inds],
                                                 scores[tracklet_inds]):
            id = int(id)
            if id in self.tracks:
                self.tracks[id]['bboxes'].append(bbox)
                self.tracks[id]['labels'].append(label)
                self.tracks[id]['scores'].append(score)
                self.tracks[id]['embeds'] = (
                    1 - self.memo_momentum
                ) * self.tracks[id]['embeds'] + self.memo_momentum * embed
                self.tracks[id]['frame_ids'].append(frame_id)
            else:
                self.tracks[id] = dict(
                    bboxes=[bbox],
                    labels=[label],
                    scores=[score],
                    embeds=embed,
                    frame_ids=[frame_id])

        # pop memo
        invalid_ids = []
        for k, v in self.tracks.items():
            if frame_id - v['frame_ids'][-1] >= self.memo_frames:
                invalid_ids.append(k)
        for invalid_id in invalid_ids:
            self.tracks.pop(invalid_id)

    @property
    def memo(self) -> Tuple[Tensor, ...]:
        """Get tracks memory."""
        memo_ids = []
        memo_bboxes = []
        memo_labels = []
        memo_scores = []
        memo_embeds = []
        for k, v in self.tracks.items():
            memo_ids.append(k)
            memo_bboxes.append(v['bboxes'][-1][None, :])
            memo_labels.append(v['labels'][-1].view(1, 1))
            memo_scores.append(v['scores'][-1].view(1, 1))
            memo_embeds.append(v['embeds'][None, :])
        memo_ids = torch.tensor(memo_ids, dtype=torch.long).view(1, -1)

        memo_bboxes = torch.cat(memo_bboxes, dim=0)
        memo_embeds = torch.cat(memo_embeds, dim=0)
        memo_labels = torch.cat(memo_labels, dim=0).squeeze(1)
        memo_scores = torch.cat(memo_scores, dim=0).squeeze(1)
        return memo_bboxes, memo_labels, memo_scores, memo_embeds, memo_ids.\
            squeeze(0)

    def track(self,
              model: torch.nn.Module,
              img: torch.Tensor,
              feats: List[torch.Tensor],
              data_sample: TrackDataSample,
              temperature: int = -1,
              rescale=True,
              **kwargs) -> InstanceData:
        """Tracking forward function.

        Args:
            model (nn.Module): MOT model.
            img (Tensor): of shape (T, C, H, W) encoding input image.
                Typically these should be mean centered and std scaled.
                The T denotes the number of key images and usually is 1 in
                QDTrack method.
            feats (list[Tensor]): Multi level feature maps of `img`.
            data_sample (:obj:`TrackDataSample`): The data sample.
                It includes information such as `pred_det_instances`.
            temperature (int): similarity temperature.
            rescale (bool, optional): If True, the bounding boxes should be
                rescaled to fit the original scale of the image. Defaults to
                True.

        Returns:
            :obj:`InstanceData`: Tracking results of the input images.
            Each InstanceData usually contains ``bboxes``, ``labels``,
            ``scores`` and ``instances_id``.
        """
        metainfo = data_sample.metainfo
        bboxes = data_sample.pred_det_instances.bboxes
        labels = data_sample.pred_det_instances.labels
        scores = data_sample.pred_det_instances.scores

        frame_id = metainfo.get('frame_id', -1)
        # create pred_track_instances
        pred_track_instances = InstanceData()
        # return zero bboxes if there is no track targets
        if bboxes.shape[0] == 0:
            ids = torch.zeros_like(labels)
            pred_track_instances = data_sample.pred_det_instances.clone()
            pred_track_instances.instances_id = ids
            return pred_track_instances

        # get track feats
        rescaled_bboxes = bboxes.clone()
        if rescale:
            scale_factor = rescaled_bboxes.new_tensor(
                metainfo['scale_factor']).repeat((1, 2))
            rescaled_bboxes = rescaled_bboxes * scale_factor
        track_feats = model.track_head.predict(feats, [rescaled_bboxes])

        # all objects is valid here
        valid_inds = labels > -1
        # inter-class nms
        low_inds = torch.nonzero(
            scores < self.distractor_score_thr, as_tuple=False).squeeze(1)
        cat_same = labels[low_inds].view(-1, 1) == labels.view(1, -1)
        ious = bbox_overlaps(bboxes[low_inds], bboxes)
        ious *= cat_same.to(ious.device)
        for i, ind in enumerate(low_inds):
            if (ious[i, :ind] > self.distractor_nms_thr).any():
                valid_inds[ind] = False
        bboxes = bboxes[valid_inds]
        scores = scores[valid_inds]
        labels = labels[valid_inds]
        embeds = track_feats[valid_inds]

        # match if buffer is not empty
        if bboxes.size(0) > 0 and not self.empty:
            memo_bboxes, memo_labels, memo_scores, memo_embeds, memo_ids \
                = self.memo

            if self.match_metric == 'bisoftmax':
                sims = embed_similarity(
                    embeds,
                    memo_embeds,
                    method='dot_product',
                    temperature=temperature)
                cat_same = labels.view(-1, 1) == memo_labels.view(1, -1)
                exps = torch.exp(sims) * cat_same.to(sims.device)
                d2t_scores = exps / (exps.sum(dim=1).view(-1, 1) + 1e-6)
                t2d_scores = exps / (exps.sum(dim=0).view(1, -1) + 1e-6)
                cos_scores = embed_similarity(
                    embeds, memo_embeds, method='cosine')
                cos_scores *= cat_same.to(cos_scores.device)
                match_scores = (d2t_scores + t2d_scores) / 2
                if self.match_with_cosine:
                    match_scores = (match_scores + cos_scores) / 2
            elif self.match_metric == 'cosine':
                cos_scores = embed_similarity(
                    embeds, memo_embeds, method='cosine')
                cat_same = labels.view(-1, 1) == memo_labels.view(1, -1)
                match_scores = cos_scores * cat_same.float().to(
                    cos_scores.device)
            else:
                raise NotImplementedError()

            # keep the object score consistency for detection of the same track
            obj_score_diffs = torch.abs(
                scores.view(-1, 1).expand_as(match_scores) -
                memo_scores.view(1, -1).expand_as(match_scores))

            num_objs = bboxes.size(0)
            ids = torch.full((num_objs, ), -1, dtype=torch.long)
            for i in range(num_objs):
                if scores[i] < self.obj_score_thr:
                    continue
                conf, memo_ind = torch.max(match_scores[i, :], dim=0)
                obj_score_diff = obj_score_diffs[i, memo_ind]
                # update track and object score for matched detection
                if (conf > self.match_score_thr) and (obj_score_diff <
                                                      self.obj_score_diff_thr):
                    ids[i] = memo_ids[memo_ind]
                    match_scores[:i, memo_ind] = 0
                    match_scores[i + 1:, memo_ind] = 0
                    m = self.momentum_obj_score
                    scores[i] = m * scores[i] + (1 - m) * memo_scores[memo_ind]
        else:
            ids = torch.full((bboxes.size(0), ), -1, dtype=torch.long)
        # init tracklets
        new_inds = (ids == -1) & (scores > self.init_score_thr).cpu()
        num_news = new_inds.sum()
        ids[new_inds] = torch.arange(
            self.num_tracks, self.num_tracks + num_news, dtype=torch.long)
        self.num_tracks += num_news

        self.update(ids, bboxes, labels, embeds, scores, frame_id)

        tracklet_inds = ids > -1
        # update pred_track_instances
        pred_track_instances.bboxes = bboxes[tracklet_inds]
        pred_track_instances.labels = labels[tracklet_inds]
        pred_track_instances.scores = scores[tracklet_inds]
        pred_track_instances.instances_id = ids[tracklet_inds]

        return pred_track_instances



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Union

import mmcv
import numpy as np
import seaborn as sns
from mmdet.structures.mask import bitmap_to_polygon
from mmdet.visualization import DetLocalVisualizer as MMDET_DetLocalVisualizer
from mmdet.visualization.palette import _get_adaptive_scales
from mmengine.dist import master_only
from mmengine.structures import InstanceData
from mmengine.visualization import Visualizer

from mmtrack.registry import VISUALIZERS
from mmtrack.structures import TrackDataSample


def random_color(seed):
    """Random a color according to the input seed."""
    np.random.seed(seed)
    colors = sns.color_palette()
    color = colors[np.random.choice(range(len(colors)))]
    color = tuple([int(255 * c) for c in color])
    return color


@VISUALIZERS.register_module()
class TrackLocalVisualizer(Visualizer):
    """MMTracking Local Visualizer for the MOT, VIS, SOT, VOS tasks.

    Args:
        name (str): Name of the instance. Defaults to 'visualizer'.
        image (np.ndarray, optional): the origin image to draw. The format
            should be RGB. Defaults to None.
        vis_backends (list, optional): Visual backend config list.
            Defaults to None.
        save_dir (str, optional): Save file dir for all storage backends.
            If it is None, the backend storage will not save any data.
        line_width (int, float): The linewidth of lines.
            Defaults to 3.
        alpha (int, float): The transparency of bboxes or mask.
                Defaults to 0.8.
    """

    def __init__(self,
                 name: str = 'visualizer',
                 image: Optional[np.ndarray] = None,
                 vis_backends: Optional[Dict] = None,
                 save_dir: Optional[str] = None,
                 line_width: Union[int, float] = 3,
                 alpha: float = 0.8):
        super().__init__(name, image, vis_backends, save_dir)
        self.line_width = line_width
        self.alpha = alpha
        # Set default value. When calling
        # `TrackLocalVisualizer().dataset_meta=xxx`,
        # it will override the default value.
        self.dataset_meta = {}

    def _draw_instances(self, image: np.ndarray,
                        instances: InstanceData) -> np.ndarray:
        """Draw instances of GT or prediction.

        Args:
            image (np.ndarray): The image to draw.
            instances (:obj:`InstanceData`): Data structure for
                instance-level annotations or predictions.

        Returns:
            np.ndarray: the drawn image which channel is RGB.
        """
        self.set_image(image)
        classes = self.dataset_meta.get('CLASSES', None)

        # get colors and texts
        if hasattr(instances, 'instances_id'):
            # for the MOT and VIS tasks
            colors = [random_color(_id) for _id in instances.instances_id]
            categories = [
                classes[label] if classes is not None else f'cls{label}'
                for label in instances.labels
            ]
            if 'scores' in instances:
                texts = [
                    f'{category_name}\n{instance_id} | {score:.2f}'
                    for category_name, instance_id, score in zip(
                        categories, instances.instances_id, instances.scores)
                ]
            else:
                texts = [
                    f'{category_name}\n{instance_id}'
                    for category_name, instance_id in zip(
                        categories, instances.instances_id)
                ]
        else:
            # for the SOT and VOS tasks
            num_instances = max(
                len(instances.get('bboxes', [])),
                len(instances.get('masks', [])))
            colors = [random_color(_) for _ in range(num_instances)]
            if 'scores' in instances:
                texts = [f'{score:.2f}' for score in instances.scores]
            else:
                texts = None

        # draw bboxes and texts
        if 'bboxes' in instances:
            # draw bboxes
            bboxes = instances.bboxes.clone()
            self.draw_bboxes(
                bboxes,
                edge_colors=colors,
                alpha=self.alpha,
                line_widths=self.line_width)
            # draw texts
            if texts is not None:
                positions = bboxes[:, :2] + self.line_width
                areas = (bboxes[:, 3] - bboxes[:, 1]) * (
                    bboxes[:, 2] - bboxes[:, 0])
                scales = _get_adaptive_scales(areas.cpu().numpy())
                for i, pos in enumerate(positions):
                    self.draw_texts(
                        texts[i],
                        pos,
                        colors='black',
                        font_sizes=int(13 * scales[i]),
                        bboxes=[{
                            'facecolor': [c / 255 for c in colors[i]],
                            'alpha': 0.8,
                            'pad': 0.7,
                            'edgecolor': 'none'
                        }])

        # draw masks
        if 'masks' in instances:
            masks = instances.masks
            polygons = []
            for i, mask in enumerate(masks):
                contours, _ = bitmap_to_polygon(mask)
                polygons.extend(contours)
            self.draw_polygons(polygons, edge_colors='w', alpha=self.alpha)
            self.draw_binary_masks(masks, colors=colors, alphas=self.alpha)

        return self.get_image()

    @master_only
    def add_datasample(
            self,
            name: str,
            image: np.ndarray,
            data_sample: Optional['TrackDataSample'] = None,
            draw_gt: bool = True,
            draw_pred: bool = True,
            show: bool = False,
            wait_time: int = 0,
            # TODO: Supported in mmengine's Viusalizer.
            out_file: Optional[str] = None,
            pred_score_thr: float = 0.3,
            step: int = 0) -> None:
        """Draw datasample and save to all backends.

        - If GT and prediction are plotted at the same time, they are
        displayed in a stitched image where the left image is the
        ground truth and the right image is the prediction.
        - If ``show`` is True, all storage backends are ignored, and
        the images will be displayed in a local window.
        - If ``out_file`` is specified, the drawn image will be
        saved to ``out_file``. t is usually used when the display
        is not available.

        Args:
            name (str): The image identifier.
            image (np.ndarray): The image to draw.
            data_sample (:obj:`TrackDataSample`, optional): A data
                sample that contain annotations and predictions.
                Defaults to None.
            draw_gt (bool): Whether to draw GT TrackDataSample.
                Default to True.
            draw_pred (bool): Whether to draw Prediction DTrackDataSample.
                Defaults to True.
            show (bool): Whether to display the drawn image. Default to False.
            wait_time (int): The interval of show (s). Defaults to 0.
            out_file (str): Path to output file. Defaults to None.
            pred_score_thr (float): The threshold to visualize the bboxes
                and masks. Defaults to 0.3.
            step (int): Global step value to record. Defaults to 0.
        """
        gt_img_data = None
        pred_img_data = None

        if data_sample is not None:
            data_sample = data_sample.cpu()

        if draw_gt and data_sample is not None:
            assert 'gt_instances' in data_sample
            gt_img_data = self._draw_instances(image, data_sample.gt_instances)

        if draw_pred and data_sample is not None:
            assert 'pred_track_instances' in data_sample
            pred_instances = data_sample.pred_track_instances
            if 'scores' in pred_instances:
                pred_instances = pred_instances[
                    pred_instances.scores > pred_score_thr].cpu()
            pred_img_data = self._draw_instances(image, pred_instances)

        if gt_img_data is not None and pred_img_data is not None:
            drawn_img = np.concatenate((gt_img_data, pred_img_data), axis=1)
        elif gt_img_data is not None:
            drawn_img = gt_img_data
        else:
            drawn_img = pred_img_data

        if show:
            self.show(drawn_img, win_name=name, wait_time=wait_time)
        else:
            self.add_image(name, drawn_img, step)

        if out_file is not None:
            mmcv.imwrite(drawn_img[..., ::-1], out_file)


@VISUALIZERS.register_module()
class DetLocalVisualizer(MMDET_DetLocalVisualizer):
    """MMTracking Local Visualizer for the VID task."""

    def add_datasample(self, *args, **kwargs):
        """Draw datasample and save to all backends."""
        if 'data_sample' in kwargs:
            # assign `pred_det_instances` to `pred_instances`
            # to hack the interface of `super().add_datasample()`.
            if 'pred_det_instances' in kwargs['data_sample']:
                kwargs['data_sample'].pred_instances = \
                    kwargs['data_sample'].pred_det_instances
        super().add_datasample(*args, **kwargs)



_base_ = './mask2former_r50_8xb2-8e_youtubevis2021.py'

model = dict(
    backbone=dict(
        depth=101,
        init_cfg=dict(type='Pretrained',
                      checkpoint='torchvision://resnet101')),
    init_cfg=dict(
        type='Pretrained',
        checkpoint='https://download.openmmlab.com/mmdetection/v2.0/'
        'mask2former/mask2former_r101_lsj_8x2_50e_coco/'
        'mask2former_r101_lsj_8x2_50e_coco_20220426_100250-c50b6fa6.pth'))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

from torch import Tensor

from mmtrack.registry import MODELS, TASK_UTILS
from mmtrack.utils import OptConfigType, SampleList
from .base import BaseMultiObjectTracker


@MODELS.register_module()
class DeepSORT(BaseMultiObjectTracker):
    """Simple online and realtime tracking with a deep association metric.

    Details can be found at `DeepSORT<https://arxiv.org/abs/1703.07402>`_.

    Args:
        detector (dict): Configuration of detector. Defaults to None.
        reid (dict): Configuration of reid. Defaults to None
        tracker (dict): Configuration of tracker. Defaults to None.
        motion (dict): Configuration of motion. Defaults to None.
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
        init_cfg (dict or list[dict]): Configuration of initialization.
            Defaults to None.
    """

    def __init__(self,
                 detector: Optional[dict] = None,
                 reid: Optional[dict] = None,
                 tracker: Optional[dict] = None,
                 motion: Optional[dict] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptConfigType = None):
        super().__init__(data_preprocessor, init_cfg)

        if detector is not None:
            self.detector = MODELS.build(detector)

        if reid is not None:
            self.reid = MODELS.build(reid)

        if motion is not None:
            self.motion = TASK_UTILS.build(motion)

        if tracker is not None:
            self.tracker = MODELS.build(tracker)

        self.preprocess_cfg = data_preprocessor

    def loss(self, inputs: Dict[str, Tensor], data_samples: SampleList,
             **kwargs) -> dict:
        """Calculate losses from a batch of inputs and data samples."""
        raise NotImplementedError(
            'Please train `detector` and `reid` models firstly, then \
                inference with SORT/DeepSORT.')

    def predict(self,
                inputs: Dict[str, Tensor],
                data_samples: SampleList,
                rescale: bool = True,
                **kwargs) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size.The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.
            rescale (bool, Optional): If False, then returned bboxes and masks
                will fit the scale of img, otherwise, returned bboxes and masks
                will fit the scale of original image shape. Defaults to True.

        Returns:
            SampleList: Tracking results of the
                input images. Each TrackDataSample usually contains
                ``pred_det_instances`` or ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'SORT/DeepSORT inference only support ' \
            '1 batch size per gpu for now.'
        img = img[0]

        assert len(data_samples) == 1, \
            'SORT/DeepSORT inference only support ' \
            '1 batch size per gpu for now.'

        track_data_sample = data_samples[0]

        det_results = self.detector.predict(img, data_samples)
        assert len(det_results) == 1, 'Batch inference is not supported.'
        track_data_sample.pred_det_instances = det_results[
            0].pred_instances.clone()

        pred_track_instances = self.tracker.track(
            model=self,
            img=img,
            feats=None,
            data_sample=track_data_sample,
            data_preprocessor=self.preprocess_cfg,
            rescale=rescale,
            **kwargs)
        track_data_sample.pred_track_instances = pred_track_instances

        return [track_data_sample]



# Copyright (c) OpenMMLab. All rights reserved.
from .base_sot_dataset import BaseSOTDataset
from .base_video_dataset import BaseVideoDataset
from .dancetrack_dataset import DanceTrackDataset
from .dataset_wrappers import RandomSampleConcatDataset
from .got10k_dataset import GOT10kDataset
from .imagenet_vid_dataset import ImagenetVIDDataset
from .lasot_dataset import LaSOTDataset
from .mot_challenge_dataset import MOTChallengeDataset
from .otb_dataset import OTB100Dataset
from .reid_dataset import ReIDDataset
from .samplers import EntireVideoBatchSampler, QuotaSampler, VideoSampler
from .sot_coco_dataset import SOTCocoDataset
from .sot_imagenet_vid_dataset import SOTImageNetVIDDataset
from .tao_dataset import TaoDataset
from .trackingnet_dataset import TrackingNetDataset
from .uav123_dataset import UAV123Dataset
from .vot_dataset import VOTDataset
from .youtube_vis_dataset import YouTubeVISDataset

__all__ = [
    'BaseVideoDataset', 'MOTChallengeDataset', 'BaseSOTDataset',
    'LaSOTDataset', 'ReIDDataset', 'GOT10kDataset', 'SOTCocoDataset',
    'SOTImageNetVIDDataset', 'TrackingNetDataset', 'YouTubeVISDataset',
    'ImagenetVIDDataset', 'RandomSampleConcatDataset', 'TaoDataset',
    'UAV123Dataset', 'VOTDataset', 'OTB100Dataset', 'DanceTrackDataset',
    'VideoSampler', 'QuotaSampler', 'EntireVideoBatchSampler'
]



_base_ = [
    './tracktor_faster-rcnn_r50-fpn_8xb2-4e_mot17halftrain'
    '_test-mot17halfval.py'
]

model = dict(
    detector=dict(
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/faster_rcnn/faster-rcnn_r50_fpn_4e_mot17-ffa52ae7.pth'  # noqa: E501
        )))

# dataloader
val_dataloader = dict(
    dataset=dict(ann_file='annotations/train_cocoformat.json', ))
test_dataloader = dict(
    dataset=dict(
        ann_file='annotations/test_cocoformat.json',
        data_prefix=dict(img_path='test'),
    ))

# evaluator
test_evaluator = dict(format_only=True, outfile_prefix='./mot_17_test_res')



_base_ = [
    '../../_base_/datasets/trackingnet.py',
    './siamese-rpn_r50_8xb28-20e_imagenetvid-imagenetdet-coco_base.py'
]

# evaluator
val_evaluator = dict(outfile_prefix='results/siamese_rpn_trackingnet')
test_evaluator = val_evaluator



_base_ = [
    '../../_base_/models/yolox_x_8x8.py',
    '../../_base_/datasets/mot_challenge.py', '../../_base_/default_runtime.py'
]

dataset_type = 'MOTChallengeDataset'
data_root = 'data/MOT17/'

img_scale = (800, 1440)
batch_size = 4

model = dict(
    type='ByteTrack',
    data_preprocessor=dict(
        type='TrackDataPreprocessor',
        pad_size_divisor=32,
        batch_augments=[
            dict(
                type='mmdet.BatchSyncRandomResize',
                random_size_range=(576, 1024),
                size_divisor=32,
                interval=10)
        ]),
    detector=dict(
        _scope_='mmdet',
        bbox_head=dict(num_classes=1),
        test_cfg=dict(score_thr=0.01, nms=dict(type='nms', iou_threshold=0.7)),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_x_8x8_300e_coco/yolox_x_8x8_300e_coco_20211126_140254-1ef88d67.pth'  # noqa: E501
        )),
    motion=dict(type='KalmanFilter'),
    tracker=dict(
        type='ByteTracker',
        obj_score_thrs=dict(high=0.6, low=0.1),
        init_track_thr=0.7,
        weight_iou_with_det_scores=True,
        match_iou_thrs=dict(high=0.1, low=0.5, tentative=0.3),
        num_frames_retain=30))

train_pipeline = [
    dict(
        type='mmdet.Mosaic',
        img_scale=img_scale,
        pad_val=114.0,
        bbox_clip_border=False),
    dict(
        type='mmdet.RandomAffine',
        scaling_ratio_range=(0.1, 2),
        border=(-img_scale[0] // 2, -img_scale[1] // 2),
        bbox_clip_border=False),
    dict(
        type='mmdet.MixUp',
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        pad_val=114.0,
        bbox_clip_border=False),
    dict(type='mmdet.YOLOXHSVRandomAug'),
    dict(type='mmdet.RandomFlip', prob=0.5),
    dict(
        type='mmdet.Resize',
        scale=img_scale,
        keep_ratio=True,
        clip_object_border=False),
    dict(
        type='mmdet.Pad',
        size_divisor=32,
        pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(
        type='mmdet.FilterAnnotations',
        min_gt_bbox_wh=(1, 1),
        keep_empty=False),
    dict(type='PackTrackInputs', pack_single_img=True)
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='mmdet.Resize', scale=img_scale, keep_ratio=True),
    dict(
        type='mmdet.Pad',
        size_divisor=32,
        pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='PackTrackInputs', pack_single_img=True)
]

train_dataloader = dict(
    _delete_=True,
    batch_size=batch_size,
    num_workers=4,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    dataset=dict(
        type='mmdet.MultiImageMixDataset',
        dataset=dict(
            type='mmdet.ConcatDataset',
            datasets=[
                dict(
                    type='mmdet.CocoDataset',
                    data_root='data/MOT17',
                    ann_file='annotations/half-train_cocoformat.json',
                    # TODO: mmdet use img as key, but img_path is needed
                    data_prefix=dict(img='train'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian')),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadTrackAnnotations'),
                    ]),
                dict(
                    type='mmdet.CocoDataset',
                    data_root='data/crowdhuman',
                    ann_file='annotations/crowdhuman_train.json',
                    data_prefix=dict(img='train'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian')),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadTrackAnnotations'),
                    ]),
                dict(
                    type='mmdet.CocoDataset',
                    data_root='data/crowdhuman',
                    ann_file='annotations/crowdhuman_val.json',
                    data_prefix=dict(img='val'),
                    filter_cfg=dict(filter_empty_gt=True, min_size=32),
                    metainfo=dict(CLASSES=('pedestrian')),
                    pipeline=[
                        dict(type='LoadImageFromFile'),
                        dict(type='LoadTrackAnnotations'),
                    ]),
            ]),
        pipeline=train_pipeline))
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        ann_file='annotations/half-val_cocoformat.json',
        data_prefix=dict(img_path='train'),
        ref_img_sampler=None,
        load_as_video=True,
        test_mode=True,
        pipeline=test_pipeline))
test_dataloader = val_dataloader
# optimizer
# default 8 gpu
lr = 0.001 / 8 * batch_size
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='SGD', lr=lr, momentum=0.9, weight_decay=5e-4, nesterov=True),
    paramwise_cfg=dict(norm_decay_mult=0., bias_decay_mult=0.))

# some hyper parameters
# training settings
total_epochs = 80
num_last_epochs = 10
resume_from = None
interval = 5

train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=total_epochs, val_interval=1)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')
# learning policy
param_scheduler = [
    dict(
        # use quadratic formula to warm up 1 epochs
        # and lr is updated by iteration
        # TODO: fix default scope in get function
        type='mmdet.QuadraticWarmupLR',
        by_epoch=True,
        begin=0,
        end=1,
        convert_to_iter_based=True),
    dict(
        # use cosine lr from 1 to 70 epoch
        type='mmdet.CosineAnnealingLR',
        eta_min=lr * 0.05,
        begin=1,
        T_max=total_epochs - num_last_epochs,
        end=total_epochs - num_last_epochs,
        by_epoch=True,
        convert_to_iter_based=True),
    dict(
        # use fixed lr during last 10 epochs
        type='mmdet.ConstantLR',
        by_epoch=True,
        factor=1,
        begin=total_epochs - num_last_epochs,
        end=total_epochs,
    )
]

custom_hooks = [
    dict(
        type='YOLOXModeSwitchHook',
        num_last_epochs=num_last_epochs,
        priority=48),
    dict(type='mmdet.SyncNormHook', priority=48),
    dict(
        type='mmdet.EMAHook',
        ema_type='mmdet.ExpMomentumEMA',
        momentum=0.0001,
        update_buffers=True,
        priority=49)
]
default_hooks = dict(checkpoint=dict(interval=1))
# evaluator
val_evaluator = dict(postprocess_tracklet_cfg=[
    dict(type='InterpolateTracklets', min_num_frames=5, max_num_frames=20)
])
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.

import copy
from typing import List, Optional, Tuple, Union

import torch
from addict import Dict
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.utils import OptConfigType, OptMultiConfig, SampleList
from ..task_modules.motion import flow_warp_feats
from .base import BaseVideoDetector


@MODELS.register_module()
class FGFA(BaseVideoDetector):
    """Flow-Guided Feature Aggregation for Video Object Detection.

    This video object detector is the implementation of `FGFA
    <https://arxiv.org/abs/1703.10025>`_.
    """

    def __init__(self,
                 detector: dict,
                 motion: dict,
                 aggregator: dict,
                 train_cfg: Optional[dict] = None,
                 test_cfg: Optional[dict] = None,
                 frozen_modules: Optional[Union[List[str], Tuple[str],
                                                str]] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super().__init__(data_preprocessor, init_cfg)
        self.detector = MODELS.build(detector)
        self.motion = MODELS.build(motion)
        self.aggregator = MODELS.build(aggregator)
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self.preprocess_cfg = data_preprocessor

        if frozen_modules is not None:
            self.freeze_module(frozen_modules)

    def loss(self, inputs: dict, data_samples: SampleList, **kwargs) -> dict:
        """
        Args:
            inputs (dict[Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size and must be 1 in FGFA method.
                The T denotes the number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'FGFA video detectors only support 1 batch size per gpu for now.'
        assert img.size(1) == 1, \
            'FGFA video detector only has 1 key image per batch.'
        img = img[0]

        ref_img = inputs['ref_img']
        assert ref_img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert ref_img.size(0) == 1, \
            'FGFA video detectors only support 1 batch size per gpu for now.'
        ref_img = ref_img[0]

        assert len(data_samples) == 1, \
            'FGFA video detectors only support 1 batch size per gpu for now.'
        metainfo = data_samples[0].metainfo

        num_ref_imgs = ref_img.size(0)
        flow_imgs = torch.cat((img.repeat(num_ref_imgs, 1, 1, 1), ref_img),
                              dim=1)
        flows = self.motion(flow_imgs, metainfo, self.preprocess_cfg)

        img_x = self.detector.extract_feat(img)
        ref_img_x = self.detector.extract_feat(ref_img)
        assert len(img_x) == len(ref_img_x)

        x = []
        for i in range(len(img_x)):
            ref_x_single = flow_warp_feats(ref_img_x[i], flows)
            agg_x_single = self.aggregator(img_x[i], ref_x_single)
            x.append(agg_x_single)

        losses = dict()

        # Two stage detector
        if hasattr(self.detector, 'roi_head'):
            # RPN forward and loss
            if self.detector.with_rpn:
                proposal_cfg = self.detector.train_cfg.get(
                    'rpn_proposal', self.detector.test_cfg.rpn)
                rpn_data_samples = copy.deepcopy(data_samples)
                # set cat_id of gt_labels to 0 in RPN
                for data_sample in rpn_data_samples:
                    data_sample.gt_instances.labels = \
                        torch.zeros_like(data_sample.gt_instances.labels)

                rpn_losses, rpn_results_list = \
                    self.detector.rpn_head.loss_and_predict(
                        x, rpn_data_samples, proposal_cfg=proposal_cfg,
                        **kwargs)
                # avoid get same name with roi_head loss
                keys = rpn_losses.keys()
                for key in keys:
                    if 'loss' in key and 'rpn' not in key:
                        rpn_losses[f'rpn_{key}'] = rpn_losses.pop(key)
                losses.update(rpn_losses)
            else:
                rpn_results_list = []
                for i in range(len(data_samples)):
                    results = InstanceData()
                    results.bboxes = data_samples[i].proposals
                    rpn_results_list.append(results)

            roi_losses = self.detector.roi_head.loss(x, rpn_results_list,
                                                     data_samples, **kwargs)
            losses.update(roi_losses)
        # Single stage detector
        elif hasattr(self.detector, 'bbox_head'):
            bbox_losses = self.detector.bbox_head.loss(x, data_samples,
                                                       **kwargs)
            losses.update(bbox_losses)
        else:
            raise TypeError('detector must has roi_head or bbox_head.')

        return losses

    def extract_feats(self, img: Tensor, ref_img: Union[Tensor, None],
                      metainfo: dict) -> List[Tensor]:
        """Extract features for `img` during testing.

        Args:
            img (Tensor): of shape (T, C, H, W) encoding input image.
                Typically these should be mean centered and std scaled.
                The T denotes the number of key images and usually is 1 in
                FGFA method.
            ref_img (Tensor | None): of shape (T, C, H, W) encoding
                reference image. Typically these should be mean centered
                and std scaled. The T denotes the number of reference images.
                There may be no reference images in some cases.
            metainfo (dict): image information dict where each
                dict has: 'img_shape', 'scale_factor', 'flip', and may also
                contain 'filename', 'ori_shape', 'pad_shape', and
                'img_norm_cfg'. For details on the values of these keys see
                `mmtrack/datasets/transforms/formatting.py:PackTrackInputs`.

        Returns:
            list[Tensor]: Multi level feature maps of `img`.
        """
        frame_id = metainfo.get('frame_id', -1)
        assert frame_id >= 0
        num_left_ref_imgs = metainfo.get('num_left_ref_imgs', -1)
        frame_stride = metainfo.get('frame_stride', -1)

        # test with adaptive stride
        if frame_stride < 1:
            if frame_id == 0:
                self.memo = Dict()
                self.memo.img = ref_img
                ref_x = self.detector.extract_feat(ref_img)
                # 'tuple' object (e.g. the output of FPN) does not support
                # item assignment
                self.memo.feats = []
                for i in range(len(ref_x)):
                    self.memo.feats.append(ref_x[i])
            x = self.detector.extract_feat(img)
        # test with fixed stride
        else:
            if frame_id == 0:
                self.memo = Dict()
                self.memo.img = ref_img
                ref_x = self.detector.extract_feat(ref_img)
                # 'tuple' object (e.g. the output of FPN) does not support
                # item assignment
                self.memo.feats = []
                # the features of img is same as ref_x[i][[num_left_ref_imgs]]
                x = []
                for i in range(len(ref_x)):
                    self.memo.feats.append(ref_x[i])
                    x.append(ref_x[i][[num_left_ref_imgs]])
            elif frame_id % frame_stride == 0:
                assert ref_img is not None
                x = []
                ref_x = self.detector.extract_feat(ref_img)
                for i in range(len(ref_x)):
                    self.memo.feats[i] = torch.cat(
                        (self.memo.feats[i], ref_x[i]), dim=0)[1:]
                    x.append(self.memo.feats[i][[num_left_ref_imgs]])
                self.memo.img = torch.cat((self.memo.img, ref_img), dim=0)[1:]
            else:
                assert ref_img is None
                x = self.detector.extract_feat(img)

        flow_imgs = torch.cat(
            (img.repeat(self.memo.img.shape[0], 1, 1, 1), self.memo.img),
            dim=1)
        flows = self.motion(flow_imgs, metainfo, self.preprocess_cfg)

        agg_x = []
        for i in range(len(x)):
            agg_x_single = flow_warp_feats(self.memo.feats[i], flows)
            if frame_stride < 1:
                agg_x_single = torch.cat((x[i], agg_x_single), dim=0)
            else:
                agg_x_single[num_left_ref_imgs] = x[i]
            agg_x_single = self.aggregator(x[i], agg_x_single)
            agg_x.append(agg_x_single)
        return agg_x

    def predict(self,
                inputs: dict,
                data_samples: SampleList,
                rescale: bool = True) -> SampleList:
        """Test without augmentation.

        Args:
            inputs (dict[Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size and must be 1 in FGFA method.
                The T denotes the number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor, Optional): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.
            rescale (bool, Optional): If False, then returned bboxes and masks
                will fit the scale of img, otherwise, returned bboxes and masks
                will fit the scale of original image shape. Defaults to True.

        Returns:
            list[obj:`TrackDataSample`]: Tracking results of the
            input images. Each TrackDataSample usually contains
            ``pred_det_instances`` or ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'FGFA video detectors only support 1 batch size per gpu for now.'
        assert img.size(1) == 1, \
            'FGFA video detector only has 1 key image per batch.'
        img = img[0]

        if 'ref_img' in inputs:
            ref_img = inputs['ref_img']
            assert ref_img.dim(
            ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
            assert ref_img.size(0) == 1, 'FGFA video detectors only support' \
                                         ' 1 batch size per gpu for now.'
            ref_img = ref_img[0]
        else:
            ref_img = None

        assert len(data_samples) == 1, \
            'FGFA video detectors only support 1 batch size per gpu for now.'
        metainfo = data_samples[0].metainfo

        x = self.extract_feats(img, ref_img, metainfo)

        track_data_sample = copy.deepcopy(data_samples[0])

        # Two stage detector
        if hasattr(self.detector, 'roi_head'):
            if not hasattr(data_samples[0], 'proposals'):
                rpn_results_list = self.detector.rpn_head.predict(
                    x, data_samples, rescale=False)
            else:
                rpn_results_list = []
                for i in range(len(data_samples)):
                    results = InstanceData()
                    results.bboxes = data_samples[i].proposals
                    rpn_results_list.append(results)

            results_list = self.detector.roi_head.predict(
                x, rpn_results_list, data_samples, rescale=rescale)
            track_data_sample.pred_det_instances = results_list[0]
        # Single stage detector
        elif hasattr(self.detector, 'bbox_head'):
            results_list = self.detector.bbox_head.predict(
                x, data_samples, rescale=rescale)

            track_data_sample.pred_det_instances = results_list[0]
        else:
            raise TypeError('detector must has roi_head or bbox_head.')

        return [track_data_sample]



_base_ = [
    '../../_base_/models/faster-rcnn_r50_fpn.py',
    '../../_base_/default_runtime.py'
]
model = dict(
    type='QDTrack',
    data_preprocessor=dict(
        _delete_=True,
        type='TrackDataPreprocessor',
        mean=[103.530, 116.280, 123.675],
        std=[1.0, 1.0, 1.0],
        # TODO: it is different from the master branch
        bgr_to_rgb=True,
        pad_size_divisor=32),
    detector=dict(
        backbone=dict(
            norm_cfg=dict(requires_grad=False),
            style='caffe',
            init_cfg=dict(
                type='Pretrained',
                checkpoint='open-mmlab://detectron2/resnet50_caffe')),
        rpn_head=dict(bbox_coder=dict(clip_border=False)),
        roi_head=dict(
            bbox_head=dict(
                loss_bbox=dict(type='L1Loss', loss_weight=1.0),
                bbox_coder=dict(clip_border=False),
                num_classes=1)),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmdetection/v2.0/faster_rcnn/'
            'faster_rcnn_r50_fpn_1x_coco-person/'
            'faster_rcnn_r50_fpn_1x_coco-person_20201216_175929-d022e227.pth'
            # noqa: E501
        )),
    track_head=dict(
        type='QuasiDenseTrackHead',
        roi_extractor=dict(
            _scope_='mmdet',
            type='SingleRoIExtractor',
            roi_layer=dict(type='RoIAlign', output_size=7, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32]),
        embed_head=dict(
            type='QuasiDenseEmbedHead',
            num_convs=4,
            num_fcs=1,
            embed_channels=256,
            norm_cfg=dict(type='GN', num_groups=32),
            loss_track=dict(type='MultiPosCrossEntropyLoss', loss_weight=0.25),
            loss_track_aux=dict(
                type='L2Loss',
                neg_pos_ub=3,
                pos_margin=0,
                neg_margin=0.1,
                hard_mining=True,
                loss_weight=1.0)),
        loss_bbox=dict(type='L1Loss', loss_weight=1.0),
        train_cfg=dict(
            assigner=dict(
                _scope_='mmdet',
                type='MaxIoUAssigner',
                pos_iou_thr=0.7,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                match_low_quality=False,
                ignore_iof_thr=-1),
            sampler=dict(
                _scope_='mmdet',
                type='CombinedSampler',
                num=256,
                pos_fraction=0.5,
                neg_pos_ub=3,
                add_gt_as_proposals=True,
                pos_sampler=dict(type='InstanceBalancedPosSampler'),
                neg_sampler=dict(type='RandomSampler')))),
    tracker=dict(
        type='QuasiDenseTracker',
        init_score_thr=0.9,
        obj_score_thr=0.5,
        match_score_thr=0.5,
        memo_tracklet_frames=30,
        memo_backdrop_frames=1,
        memo_momentum=0.8,
        nms_conf_thr=0.5,
        nms_backdrop_iou_thr=0.3,
        nms_class_iou_thr=0.7,
        with_cats=True,
        match_metric='bisoftmax'))
# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.02, momentum=0.9, weight_decay=0.0001),
    clip_grad=dict(max_norm=35, norm_type=2))
# learning policy
param_scheduler = [
    dict(
        type='mmdet.MultiStepLR',
        begin=0,
        end=4,
        by_epoch=True,
        milestones=[3])
]

# runtime settings
train_cfg = dict(type='EpochBasedTrainLoop', max_epochs=4, val_interval=1)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from .filter_head import FilterInitializer
from .filter_optimizer import PrDiMPFilterOptimizer

__all__ = ['FilterInitializer', 'PrDiMPFilterOptimizer']



_base_ = ['../../_base_/default_runtime.py']

randomness = dict(seed=1, deterministic=True)

# model settings
model = dict(
    type='Stark',
    data_preprocessor=dict(
        type='TrackDataPreprocessor',
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        bgr_to_rgb=True),
    backbone=dict(
        type='mmdet.ResNet',
        depth=50,
        num_stages=3,
        strides=(1, 2, 2),
        dilations=[1, 1, 1],
        out_indices=[2],
        frozen_stages=1,
        norm_eval=True,
        norm_cfg=dict(type='BN', requires_grad=False),
        init_cfg=dict(type='Pretrained', checkpoint='torchvision://resnet50')),
    neck=dict(
        type='mmdet.ChannelMapper',
        in_channels=[1024],
        out_channels=256,
        kernel_size=1,
        act_cfg=None),
    head=dict(
        type='StarkHead',
        num_querys=1,
        transformer=dict(
            type='StarkTransformer',
            encoder=dict(
                type='mmdet.DetrTransformerEncoder',
                num_layers=6,
                transformerlayers=dict(
                    type='BaseTransformerLayer',
                    attn_cfgs=[
                        dict(
                            type='MultiheadAttention',
                            embed_dims=256,
                            num_heads=8,
                            attn_drop=0.1,
                            dropout_layer=dict(type='Dropout', drop_prob=0.1))
                    ],
                    ffn_cfgs=dict(
                        feedforward_channels=2048,
                        embed_dims=256,
                        ffn_drop=0.1),
                    operation_order=('self_attn', 'norm', 'ffn', 'norm'))),
            decoder=dict(
                type='mmdet.DetrTransformerDecoder',
                return_intermediate=False,
                num_layers=6,
                transformerlayers=dict(
                    type='BaseTransformerLayer',
                    attn_cfgs=dict(
                        type='MultiheadAttention',
                        embed_dims=256,
                        num_heads=8,
                        attn_drop=0.1,
                        dropout_layer=dict(type='Dropout', drop_prob=0.1)),
                    ffn_cfgs=dict(
                        feedforward_channels=2048,
                        embed_dims=256,
                        ffn_drop=0.1),
                    operation_order=('self_attn', 'norm', 'cross_attn', 'norm',
                                     'ffn', 'norm'))),
        ),
        positional_encoding=dict(
            type='mmdet.SinePositionalEncoding', num_feats=128,
            normalize=True),
        bbox_head=dict(
            type='CornerPredictorHead',
            inplanes=256,
            channel=256,
            feat_size=20,
            stride=16),
        loss_bbox=dict(type='mmdet.L1Loss', loss_weight=5.0),
        loss_iou=dict(type='mmdet.GIoULoss', loss_weight=2.0)),
    test_cfg=dict(
        search_factor=5.0,
        search_size=320,
        template_factor=2.0,
        template_size=128,
        num_templates=2))

data_root = 'data/'
train_pipeline = [
    dict(
        type='TridentSampling',
        num_search_frames=1,
        num_template_frames=2,
        max_frame_range=[200],
        cls_pos_prob=0.5,
        train_cls_head=False),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=False),
            dict(type='GrayAug', prob=0.05),
            dict(type='mmdet.RandomFlip', prob=0.5, direction='horizontal')
        ]),
    dict(
        type='SeqBboxJitter',
        center_jitter_factor=[0, 0, 4.5],
        scale_jitter_factor=[0, 0, 0.5],
        crop_size_factor=[2, 2, 5]),
    dict(
        type='SeqCropLikeStark',
        crop_size_factor=[2, 2, 5],
        output_size=[128, 128, 320]),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[dict(type='BrightnessAug', jitter_range=0.2)]),
    dict(type='CheckPadMaskValidity', stride=16),
    dict(type='PackTrackInputs', ref_prefix='search', num_template_frames=2)
]

# dataset settings
train_dataloader = dict(
    batch_size=16,
    num_workers=4,
    persistent_workers=True,
    sampler=dict(type='QuotaSampler', samples_per_epoch=60000),
    dataset=dict(
        type='RandomSampleConcatDataset',
        dataset_sampling_weights=[1, 1, 1, 1],
        datasets=[
            dict(
                type='GOT10kDataset',
                data_root=data_root,
                ann_file='GOT10k/annotations/got10k_train_vot_infos.txt',
                data_prefix=dict(img_path='GOT10k'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='LaSOTDataset',
                data_root=data_root,
                ann_file='LaSOT_full/annotations/lasot_train_infos.txt',
                data_prefix=dict(img_path='LaSOT_full/LaSOTBenchmark'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='TrackingNetDataset',
                data_root=data_root,
                ann_file='TrackingNet/annotations/trackingnet_train_infos.txt',
                data_prefix=dict(img_path='TrackingNet'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='SOTCocoDataset',
                data_root=data_root,
                ann_file='coco/annotations/instances_train2017.json',
                data_prefix=dict(img_path='coco/train2017'),
                pipeline=train_pipeline,
                test_mode=False)
        ]))

# runner loop
train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=500, val_begin=500, val_interval=1)

# learning policy
param_scheduler = dict(type='MultiStepLR', milestones=[400], gamma=0.1)

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='AdamW', lr=0.0001, weight_decay=0.0001),
    clip_grad=dict(max_norm=0.1, norm_type=2),
    paramwise_cfg=dict(
        custom_keys=dict(backbone=dict(lr_mult=0.1, decay_mult=1.0))))

# checkpoint saving
default_hooks = dict(checkpoint=dict(type='CheckpointHook', interval=100))



_base_ = [
    '../../_base_/models/yolox_x_8x8.py',
    '../../_base_/datasets/mot_challenge.py', '../../_base_/default_runtime.py'
]

model = dict(
    type='StrongSORT',
    data_preprocessor=dict(
        type='TrackDataPreprocessor',
        pad_size_divisor=32,
        batch_augments=[
            dict(
                type='mmdet.BatchSyncRandomResize',
                random_size_range=(480, 800),
                size_divisor=32,
                interval=10)
        ]),
    detector=dict(
        _scope_='mmdet',
        bbox_head=dict(num_classes=1),
        test_cfg=dict(score_thr=0.01, nms=dict(type='nms', iou_threshold=0.7)),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/strongsort/mot_dataset/yolox_x_crowdhuman_mot17-private-half_20220812_192036-b6c9ce9a.pth'  # noqa: E501
        )),
    kalman=dict(type='KalmanFilter', center_only=False, use_nsa=True),
    cmc=dict(
        type='CameraMotionCompensation',
        warp_mode='cv2.MOTION_EUCLIDEAN',
        num_iters=100,
        stop_eps=0.00001),
    reid=dict(
        type='BaseReID',
        data_preprocessor=None,
        backbone=dict(
            type='mmcls.ResNet',
            depth=50,
            num_stages=4,
            out_indices=(3, ),
            style='pytorch'),
        neck=dict(type='GlobalAveragePooling', kernel_size=(8, 4), stride=1),
        head=dict(
            type='LinearReIDHead',
            num_fcs=1,
            in_channels=2048,
            fc_channels=1024,
            out_channels=128,
            num_classes=380,
            loss_cls=dict(type='mmcls.CrossEntropyLoss', loss_weight=1.0),
            loss_triplet=dict(type='TripletLoss', margin=0.3, loss_weight=1.0),
            norm_cfg=dict(type='BN1d'),
            act_cfg=dict(type='ReLU')),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/reid/reid_r50_6e_mot17-4bf6b63d.pth'  # noqa: E501
        )),
    tracker=dict(
        type='StrongSORTTracker',
        obj_score_thr=0.6,
        reid=dict(
            num_samples=None,
            img_scale=(256, 128),
            img_norm_cfg=dict(
                mean=[123.675, 116.28, 103.53],
                std=[58.395, 57.12, 57.375],
                to_rgb=True),
            match_score_thr=0.3,
            motion_weight=0.02,
        ),
        match_iou_thr=0.7,
        momentums=dict(embeds=0.1, ),
        num_tentatives=2,
        num_frames_retain=100))

dataset_type = 'MOTChallengeDataset'
data_root = 'data/MOT17/'
img_scale = (800, 1440)

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='mmdet.Resize', scale=img_scale, keep_ratio=True),
    dict(
        type='mmdet.Pad',
        size_divisor=32,
        pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='PackTrackInputs', pack_single_img=True)
]

train_dataloader = None
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        ann_file='annotations/half-val_cocoformat.json',
        data_prefix=dict(img_path='train'),
        ref_img_sampler=None,
        load_as_video=True,
        test_mode=True,
        pipeline=test_pipeline))
test_dataloader = val_dataloader

train_cfg = None
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# evaluator
val_evaluator = dict(postprocess_tracklet_cfg=[
    dict(
        type='AppearanceFreeLink',
        checkpoint=  # noqa: E251
        'https://download.openmmlab.com/mmtracking/mot/strongsort/mot_dataset/aflink_motchallenge_20220812_190310-a7578ad3.pth',  # noqa: E501
        temporal_threshold=(0, 30),
        spatial_threshold=50,
        confidence_threshold=0.95,
    ),
    dict(
        type='InterpolateTracklets',
        min_num_frames=5,
        max_num_frames=20,
        use_gsi=True,
        smooth_tau=10)
])
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import List

import numpy as np
from mmengine.dataset import force_full_init
from mmengine.fileio.file_client import FileClient
from pycocotools.coco import COCO

from mmtrack.registry import DATASETS
from .base_sot_dataset import BaseSOTDataset


@DATASETS.register_module()
class SOTCocoDataset(BaseSOTDataset):
    """Coco dataset of single object tracking.

    The dataset only support training mode.
    """

    def __init__(self, *args, **kwargs):
        """Initialization of SOT dataset class."""
        super().__init__(*args, **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load annotations from an annotation file named as ``self.ann_file``.
        Each instance is viewed as a video.

        Returns:
            list[dict]: The length of the list is the number of valid object
                annotations. The inner dict contains annotation ID in coco
                API.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_file:
            self.coco = COCO(local_file)
        ann_list = list(self.coco.anns.keys())
        data_infos = [
            dict(ann_id=ann) for ann in ann_list
            if self.coco.anns[ann].get('iscrowd', 0) == 0
        ]
        return data_infos

    def get_bboxes_from_video(self, video_idx: int) -> np.ndarray:
        """Get bbox annotation about one instance in an image.

        Args:
            video_idx (int): The index of video.

        Returns:
            ndarray: In [1, 4] shape. The bbox is in (x, y, w, h) format.
        """
        ann_id = self.get_data_info(video_idx)['ann_id']
        anno = self.coco.anns[ann_id]
        bboxes = np.array(anno['bbox'], dtype=np.float32).reshape(-1, 4)
        return bboxes

    def get_img_infos_from_video(self, video_idx: int) -> dict:
        """Get the image information about one instance in a image.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: {
                    'video_id': int,
                    'frame_ids': np.ndarray,
                    'img_paths': list[str],
                    'video_length': int
                  }
        """
        ann_id = self.get_data_info(video_idx)['ann_id']
        imgs = self.coco.loadImgs([self.coco.anns[ann_id]['image_id']])
        img_names = [
            osp.join(self.data_prefix['img_path'], img['file_name'])
            for img in imgs
        ]
        frame_ids = np.arange(self.get_len_per_video(video_idx))
        img_infos = dict(
            video_id=video_idx,
            frame_ids=frame_ids,
            img_paths=img_names,
            video_length=1)
        return img_infos

    @force_full_init
    def get_len_per_video(self, video_idx: int) -> int:
        """Get the number of frames in a video. Here, it returns 1 since Coco
        is a image dataset.

        Args:
            video_idx (int): The index of video. Each video_idx denotes an
                instance.

        Returns:
            int: The length of video.
        """
        return 1



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Tuple

from mmdet.models import StandardRoIHead
from mmdet.structures.bbox import bbox2roi
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.utils import ConfigType, InstanceList, SampleList


@MODELS.register_module()
class SelsaRoIHead(StandardRoIHead):
    """selsa roi head."""

    def loss(self, x: Tuple[Tensor], ref_x: Tuple[Tensor],
             rpn_results_list: InstanceList,
             ref_rpn_results_list: InstanceList,
             data_samples: SampleList) -> dict:
        """
        Args:
            x (Tuple[Tensor]): list of multi-level img features.
            ref_x (Tuple[Tensor]): list of multi-level ref_img features.
            rpn_results_list (InstanceList): list of region proposals.
            ref_rpn_results_list (InstanceList): list of region proposals
                from reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance` and 'metainfo'.

        Returns:
            dict[str, Tensor]: a dictionary of loss components.
        """
        assert len(rpn_results_list) == len(data_samples)
        batch_gt_instances = []
        batch_gt_instances_ignore = []
        for data_sample in data_samples:
            batch_gt_instances.append(data_sample.gt_instances)
            if 'ignored_instances' in data_sample:
                batch_gt_instances_ignore.append(data_sample.ignored_instances)
            else:
                batch_gt_instances_ignore.append(None)

        # assign gts and sample proposals
        if self.with_bbox or self.with_mask:
            num_imgs = len(data_samples)
            sampling_results = []
            for i in range(num_imgs):
                # rename rpn_results.bboxes to rpn_results.priors
                rpn_results = rpn_results_list[i]
                rpn_results.priors = rpn_results.pop('bboxes')
                assign_result = self.bbox_assigner.assign(
                    rpn_results, batch_gt_instances[i],
                    batch_gt_instances_ignore[i])
                sampling_result = self.bbox_sampler.sample(
                    assign_result,
                    rpn_results,
                    batch_gt_instances[i],
                    feats=[lvl_feat[i][None] for lvl_feat in x])
                sampling_results.append(sampling_result)

        losses = dict()
        # bbox head forward and loss
        if self.with_bbox:
            bbox_results = self.bbox_loss(x, ref_x, sampling_results,
                                          ref_rpn_results_list)
            losses.update(bbox_results['loss_bbox'])

        # mask head forward and loss
        if self.with_mask:
            mask_results = self.mask_loss(x, sampling_results,
                                          bbox_results['bbox_feats'],
                                          batch_gt_instances)
            losses.update(mask_results['loss_mask'])

        return losses

    def _bbox_forward(self, x: Tuple[Tensor], ref_x: Tuple[Tensor],
                      rois: Tensor, ref_rois: Tensor) -> dict:
        """Box head forward function used in both training and testing.

        Args:
            x (Tuple[Tensor]): List of multi-level img features.
            ref_x (Tuple[Tensor]): List of multi-level reference img features.
            rois (Tensor): RoIs with the shape (n, 5) where the first
                column indicates batch id of each RoI.
            ref_rois (Tensor): Reference RoIs with the shape (n, 5) where the
                first column indicates batch id of each RoI.

        Returns:
            dict[str, Tensor]: Usually returns a dictionary with keys:

                - `cls_score` (Tensor): Classification scores.
                - `bbox_pred` (Tensor): Box energies / deltas.
                - `bbox_feats` (Tensor): Extract bbox RoI features.
        """

        # TODO: a more flexible way to decide which feature maps to use
        bbox_feats = self.bbox_roi_extractor(
            x[:self.bbox_roi_extractor.num_inputs],
            rois,
            ref_feats=ref_x[:self.bbox_roi_extractor.num_inputs])

        ref_bbox_feats = self.bbox_roi_extractor(
            ref_x[:self.bbox_roi_extractor.num_inputs], ref_rois)
        if self.with_shared_head:
            bbox_feats = self.shared_head(bbox_feats)
            ref_bbox_feats = self.shared_head(ref_bbox_feats)
        cls_score, bbox_pred = self.bbox_head(bbox_feats, ref_bbox_feats)

        bbox_results = dict(
            cls_score=cls_score, bbox_pred=bbox_pred, bbox_feats=bbox_feats)
        return bbox_results

    def bbox_loss(self, x: Tuple[Tensor], ref_x: Tuple[Tensor],
                  sampling_results: InstanceList,
                  ref_rpn_results_list: InstanceList):
        """Run forward function and calculate loss for box head in training.

        Args:
            x (Tuple[Tensor]): list of multi-level img features.
            ref_x (Tuple[Tensor]): list of multi-level ref_img features.
            sampling_results (InstanceList): Sampleing results.
            ref_rpn_results_list (InstanceList): list of region proposals
                from reference images.

        Returns:
            dict[str, Tensor]: a dictionary of loss components.
        """
        rois = bbox2roi([res.bboxes for res in sampling_results])
        ref_rois = bbox2roi([res.bboxes for res in ref_rpn_results_list])
        bbox_results = self._bbox_forward(x, ref_x, rois, ref_rois)

        bbox_loss_and_target = self.bbox_head.loss_and_target(
            cls_score=bbox_results['cls_score'],
            bbox_pred=bbox_results['bbox_pred'],
            rois=rois,
            sampling_results=sampling_results,
            rcnn_train_cfg=self.train_cfg)

        bbox_results.update(loss_bbox=bbox_loss_and_target['loss_bbox'])
        return bbox_results

    def predict(self,
                x: Tuple[Tensor],
                ref_x: Tuple[Tensor],
                rpn_results_list: InstanceList,
                ref_rpn_results_list: InstanceList,
                data_samples: SampleList,
                rescale: bool = False) -> InstanceList:
        """Perform forward propagation of the roi head and predict detection
        results on the features of the upstream network.

        Args:
            x (Tuple[Tensor]): All scale level feature maps of images.
            ref_x (Tuple[Tensor]): All scale level feature maps of reference
                mages.
            rpn_results_list (InstanceList): list of region proposals.
            ref_rpn_results_list (InstanceList): list of region
                proposals from reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance` and 'metainfo'.
            rescale (bool, optional): If True, return boxes in original image
                space. Defaults to False.

        Returns:
            list[:obj:`InstanceData`]: Detection results of each image
            after the post process.
            Each item usually contains following keys.

                - scores (Tensor): Classification scores, has a shape
                  (num_instance, )
                - labels (Tensor): Labels of bboxes, has a shape
                  (num_instances, ).
                - bboxes (Tensor): Has a shape (num_instances, 4),
                  the last dimension 4 arrange as (x1, y1, x2, y2).
        """
        assert self.with_bbox, 'Bbox head must be implemented.'

        batch_img_metas = [
            data_samples.metainfo for data_samples in data_samples
        ]

        results_list = self.predict_bbox(
            x,
            ref_x,
            rpn_results_list,
            ref_rpn_results_list,
            batch_img_metas,
            self.test_cfg,
            rescale=rescale)

        if self.with_mask:
            results_list = self.predict_mask(
                x, batch_img_metas, results_list, rescale=rescale)

        return results_list

    def predict_bbox(self,
                     x: Tuple[Tensor],
                     ref_x: Tuple[Tensor],
                     rpn_results_list: InstanceList,
                     ref_rpn_results_list: InstanceList,
                     batch_img_metas: List[dict],
                     rcnn_test_cfg: ConfigType,
                     rescale: bool = False) -> InstanceList:
        """Perform forward propagation of the bbox head and predict detection
        results on the features of the upstream network.

        Args:
            x (Tuple[Tensor]): All scale level feature maps of images.
            ref_x (Tuple[Tensor]): All scale level feature maps of reference
                mages.
            rpn_results_list (InstanceList): List of region proposals.
            ref_rpn_results_list (InstanceList): List of region
                proposals from reference images.
            batch_img_metas (List[dict]): _List of image information.
            rcnn_test_cfg (ConfigType): `test_cfg` of R-CNN.
            rescale (bool, optional): If True, return boxes in original image
                space. Defaults to False.

        Returns:
            list[obj:`InstanceData`]: Detection results of each image.
            Each item usually contains following keys.

                - scores (Tensor): Classification scores, has a shape
                  (num_instance, )
                - labels (Tensor): Labels of bboxes, has a shape
                  (num_instances, ).
                - bboxes (Tensor): Has a shape (num_instances, 4),
                  the last dimension 4 arrange as (x1, y1, x2, y2).
                - masks (Tensor): Has a shape (num_instances, H, W).
        """

        rois = bbox2roi([res.bboxes for res in rpn_results_list])
        ref_rois = bbox2roi([res.bboxes for res in ref_rpn_results_list])
        bbox_results = self._bbox_forward(x, ref_x, rois, ref_rois)

        # split batch bbox prediction back to each image
        cls_score = bbox_results['cls_score']
        bbox_pred = bbox_results['bbox_pred']
        num_proposals_per_img = tuple(len(p) for p in rpn_results_list)
        rois = rois.split(num_proposals_per_img, 0)
        cls_score = cls_score.split(num_proposals_per_img, 0)
        # some detector with_reg is False, bbox_pred will be None
        bbox_pred = bbox_pred.split(
            num_proposals_per_img,
            0) if bbox_pred is not None else [None, None]

        # apply bbox post-processing to each image individually
        result_list = self.bbox_head.predict_by_feat(
            rois,
            cls_score,
            bbox_pred,
            batch_img_metas=batch_img_metas,
            rcnn_test_cfg=rcnn_test_cfg,
            rescale=rescale)

        return result_list



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Optional

from mmengine.optim.scheduler.lr_scheduler import LRSchedulerMixin
from mmengine.optim.scheduler.param_scheduler import INF, _ParamScheduler
from torch.optim import Optimizer

from mmtrack.registry import PARAM_SCHEDULERS


@PARAM_SCHEDULERS.register_module()
class SiamRPNExpParamScheduler(_ParamScheduler):
    """Decays the parameter value of each parameter group by exponentially
    changing small multiplicative factor until the number of epoch reaches a
    pre-defined milestone: ``end``.

    Notice that such decay can happen simultaneously with other changes to the
    parameter value from outside this scheduler.

    .. math::

        X_{t} = X_{t-1} \times (\frac{end}{begin})^{\frac{1}{epochs}}


    Args:
        optimizer (Optimizer): Wrapped optimizer.
        param_name (str): Name of the parameter to be adjusted, such as
            ``lr``, ``momentum``.
        start_factor (float): The number we multiply parameter value in the
            first epoch. The multiplication factor changes towards end_factor
            in the following epochs. Defaults to 0.1.
        end_factor (float): The number we multiply parameter value at the end
            of linear changing process. Defaults to 1.0.
        begin (int): Step at which to start updating the parameters.
            Defaults to 0.
        end (int): Step at which to stop updating the parameters.
            Defaults to INF.
        endpoint (bool): If true, `end_factor`` is included in the ``end``.
            Otherwise, it is not included. Default is True.
        last_step (int): The index of last step. Used for resume without
            state dict. Defaults to -1.
        by_epoch (bool): Whether the scheduled parameters are updated by
            epochs. Defaults to True.
        verbose (bool): Whether to print the value for each update.
            Defaults to False.
    """

    def __init__(self,
                 optimizer: Optimizer,
                 param_name: str,
                 start_factor: float = 0.1,
                 end_factor: float = 1.0,
                 begin: int = 0,
                 end: int = INF,
                 endpoint: bool = True,
                 last_step: int = -1,
                 by_epoch: bool = True,
                 verbose: bool = False):
        if end >= INF:
            raise ValueError('``end`` must be less than infinity,'
                             'Please set ``end`` parameter of '
                             '``QuadraticWarmupScheduler`` as the '
                             'number of warmup end.')

        if start_factor > 1.0 or start_factor < 0:
            raise ValueError(
                'Starting multiplicative factor should between 0 and 1.')

        if end_factor > 1.0 or end_factor < 0:
            raise ValueError(
                'Ending multiplicative factor should between 0 and 1.')

        self.start_factor = start_factor
        self.end_factor = end_factor
        self.endpoint = endpoint
        self.total_iters = end - begin - 1 if self.endpoint else end - begin
        super().__init__(
            optimizer,
            param_name=param_name,
            begin=begin,
            end=end,
            last_step=last_step,
            by_epoch=by_epoch,
            verbose=verbose)

    @classmethod
    def build_iter_from_epoch(cls,
                              *args,
                              begin: int = 0,
                              end: int = INF,
                              by_epoch: bool = True,
                              epoch_length: Optional[int] = None,
                              **kwargs):
        """Build an iter-based instance of this scheduler from an epoch-based
        config.

        Args:
            begin (int, optional): Step at which to start updating the
                parameters. Defaults to 0.
            end (int, optional): Step at which to stop updating the parameters.
                Defaults to INF.
            by_epoch (bool, optional): Whether the scheduled parameters are
                updated by epochs. Defaults to True.
            epoch_length (Optional[int], optional): The length of each epoch.
                Defaults to None.

        Returns:
            Object: The instantiated object of ``SiamRPNExpParamScheduler``.
        """

        assert by_epoch, 'Only epoch-based kwargs whose `by_epoch=True` can ' \
                         'be converted to iter-based.'
        assert epoch_length is not None and epoch_length > 0, \
            f'`epoch_length` must be a positive integer, ' \
            f'but got {epoch_length}.'
        by_epoch = False
        begin = begin * epoch_length
        if end != INF:
            end = end * epoch_length
        return cls(*args, begin=begin, end=end, by_epoch=by_epoch, **kwargs)

    def _get_value(self):
        """Compute value using chainable form of the scheduler."""
        if self.last_step == 0:
            return [
                group[self.param_name] * self.start_factor
                for group in self.optimizer.param_groups
            ]

        return [
            group[self.param_name] *
            math.pow(self.end_factor / self.start_factor, 1 /
                     (self.total_iters))
            for group in self.optimizer.param_groups
        ]


@PARAM_SCHEDULERS.register_module()
class SiamRPNExpLR(LRSchedulerMixin, SiamRPNExpParamScheduler):
    """Decays the parameter value of each parameter group by exponentially
    changing small multiplicative factor until the number of epoch reaches a
    pre-defined milestone: ``end``.

    Notice that such decay can happen simultaneously with other changes to the
    parameter value from outside this scheduler.

    .. math::

        X_{t} = X_{t-1} \times (\frac{end}{begin})^{\frac{1}{epochs}}


    Args:
        optimizer (Optimizer): Wrapped optimizer.
        start_factor (float): The number we multiply parameter value in the
            first epoch. The multiplication factor changes towards end_factor
            in the following epochs. Defaults to 0.1.
        end_factor (float): The number we multiply parameter value at the end
            of linear changing process. Defaults to 1.0.
        begin (int): Step at which to start updating the parameters.
            Defaults to 0.
        end (int): Step at which to stop updating the parameters.
            Defaults to INF.
        endpoint (bool): If true, `end_factor`` is included in the ``end``.
            Otherwise, it is not included. Default is True.
        last_step (int): The index of last step. Used for resume without
            state dict. Defaults to -1.
        by_epoch (bool): Whether the scheduled parameters are updated by
            epochs. Defaults to True.
        verbose (bool): Whether to print the value for each update.
            Defaults to False.
    """



_base_ = [
    '../../_base_/datasets/otb100.py',
    './siamese-rpn_r50_8xb28-20e_imagenetvid-imagenetdet-coco_base.py'
]

crop_size = 511
exemplar_size = 127
search_size = 255

# model settings
model = dict(
    test_cfg=dict(rpn=dict(penalty_k=0.4, window_influence=0.5, lr=0.4)))

data_root = {{_base_.data_root}}
train_pipeline = [
    dict(
        type='PairSampling',
        frame_range=100,
        pos_prob=0.8,
        filter_template_img=False),
    dict(
        type='TransformBroadcaster',
        share_random_params=False,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=False),
            dict(
                type='CropLikeSiamFC',
                context_amount=0.5,
                exemplar_size=exemplar_size,
                crop_size=crop_size)
        ]),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=dict(type='GrayAug', prob=0.2)),
    dict(
        type='SeqShiftScaleAug',
        target_size=[exemplar_size, search_size],
        shift=[4, 64],
        scale=[0.05, 0.18]),
    dict(type='SeqColorAug', prob=[1.0, 1.0]),
    dict(type='SeqBlurAug', prob=[0.0, 0.2]),
    dict(type='PackTrackInputs', ref_prefix='search', num_template_frames=1)
]

# dataloader
train_dataloader = dict(
    batch_size=16,
    dataset=dict(datasets=[
        dict(
            type='SOTImageNetVIDDataset',
            data_root=data_root,
            ann_file='ILSVRC/annotations/imagenet_vid_train.json',
            data_prefix=dict(img_path='ILSVRC/Data/VID'),
            pipeline=train_pipeline,
            test_mode=False),
        dict(
            type='SOTCocoDataset',
            data_root=data_root,
            ann_file='coco/annotations/instances_train2017.json',
            data_prefix=dict(img_path='coco/train2017'),
            pipeline=train_pipeline,
            test_mode=False),
        dict(
            type='SOTCocoDataset',
            data_root=data_root,
            ann_file='ILSVRC/annotations/imagenet_det_30plus1cls.json',
            data_prefix=dict(img_path='ILSVRC/Data/DET'),
            pipeline=train_pipeline,
            test_mode=False)
    ]))
